from nicegui import ui, events, app
from typing import Optional, Union, Any
from numpy import ndarray
from pandas import DataFrame

from ..type_view import TypeView

from SIMULTAN.Data.Components import (SimEnumParameter, SimIntegerParameter, SimStringParameter, SimDoubleParameter)

from ... import user_manager
from ...core.geo_associations import GeoEditDialog
from ...core.edit_dialog import ContentEditDialog

from ..component_detail_base_view import ComponentDetailBaseView

from PySimultan2.simultan_object import SimultanObject
from PySimultan2.files import FileInfo
from PySimultan2.taxonomy_maps import Content
from PySimultan2.geometry import GeometryModel


from ..parameter_view import ParameterView


class ContentItemView(object):
    def __init__(self, *args, **kwargs):
        self.component: SimultanObject = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.content = kwargs.get('content')

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):
        with ui.item().classes('w-full h-full'):
            with ui.item_section():
                ui.label(f'{self.content.name}:')

            val = getattr(self.component, self.content.property_name)
            if isinstance(val, SimultanObject):
                ui_simultan_object_ref(val, self)
            elif isinstance(val, (int, float, str)):
                with ui.item_section():
                    raw_val = self.component.get_raw_attr(self.content.property_name)
                    ParameterView(component=val,
                                  raw_val=raw_val,
                                  content=self.content,
                                  parent=self.parent,
                                  taxonomy=self.content).ui_content()

            elif isinstance(val, (ndarray, DataFrame)):
                ui_ndarray_ref(val, self)
            else:
                if val is None:
                    with ui.item_section():
                        ui.label('None')
                with ui.item_section():
                    if hasattr(val, 'name'):
                        ui.label(f'{val.name}:')
                    else:
                        ui.label('No Name')
                with ui.item_section():
                    if hasattr(val, 'id'):
                        ui.label(f'{val.id}:')
                    else:
                        ui.label('No ID')
                with ui.item_section():
                    button = ui.button(on_click=self.edit,
                                       icon='edit').classes('q-ml-auto')
                    button.item = val
                    button.content = self.content

    def edit(self, event):
        edit_dialog = ContentEditDialog(component=event.sender.item,
                                        parent=self.parent,
                                        content=event.sender.content)
        edit_dialog.create_edit_dialog()


def ui_simultan_object_ref(val: SimultanObject,
                           parent: ContentItemView):
    from ..detail_views import show_detail

    # show_detail(val)

    # if not hasattr(val, '__ui_element__') or val.__ui_element__ is None:
    #
    #     view_manager = user_manager[app.storage.user['username']].project_manager.view_manager
    #     cls_view = view_manager.cls_views.get(val.__class__, None)[
    #         'item_view_manager'] if view_manager.cls_views.get(val.__class__,
    #                                                            None) is not None else None
    #     if cls_view is None:
    #         cls_view: TypeViewManager = view_manager.create_mapped_cls_view_manager(taxonomy=val.__class__._taxonomy)[
    #             'item_view_manager']
    #         if cls_view is None:
    #             ui.label(f'No View for this class: {val.__class__}')
    #             return
    #     try:
    #         cls_view.add_item_to_view(val)
    #     except KeyError:
    #         with ui.item_section():
    #             ui.label(f'No View for this class: {val.__class__}')
    # if val.__ui_element__ is None:
    #     ui.label(f'No View for this object: {str(val), val.__class__}')
    #     return

    with ui.item_section():
        ui.label(val.name)
    with ui.item_section():
        ui.button(on_click=lambda e: show_detail(value=val,
                                                 parent=parent),
                  icon='launch').classes('q-ml-auto')
    with ui.item_section():
        with ui.row():
            ui.label(f'{val.id.GlobalId.ToString()}')
        with ui.row():
            ui.label(f'{val.id.LocalId}')
    with ui.item_section():
        edit_btn = ui.button(on_click=parent.edit, icon='edit').classes('q-ml-auto')
        edit_btn.item = val
        edit_btn.content = parent.content


def ui_ndarray_ref(val: (ndarray, DataFrame),
                   parent: ContentItemView):
    from ..detail_views import show_detail

    raw_val = parent.component.get_raw_attr(parent.content.property_name)
    # view_manager = user_manager[app.storage.user['username']].project_manager.view_manager
    #
    # def get_ui_element():
    #     ui_element = view_manager.cls_views[val.__class__]['item_view_manager'].item_views.get(
    #         str(raw_val.ValueSource.ValueField.Id), None)
    #     return ui_element
    #
    # ui_element = get_ui_element()
    #
    # if ui_element is None:
    #     view_manager.cls_views[val.__class__]['item_view_manager'].add_item_to_view(val, raw_val)
    # ui_element = get_ui_element()

    with ui.item_section():

        ui.label(raw_val.ValueSource.ValueField.Name)
        ui.button(on_click=lambda e: show_detail(value=val,
                                                 parent=parent),
                  icon='launch').classes('q-ml-auto')
    with ui.item_section():
        ui.label(f'{raw_val.ValueSource.ValueField.Id}')
    with ui.item_section():
        edit_btn = ui.button(on_click=parent.edit, icon='edit').classes('q-ml-auto')
        edit_btn.item = val
        edit_btn.content = parent.content


class ContentView(object):

    def __init__(self, *args, **kwargs):
        self.component: SimultanObject = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.card = None
        self.row = None

    @property
    def parameters(self) -> tuple[list[tuple[Content,
    Union[int, float, str],
    Union[SimIntegerParameter,
    SimStringParameter,
    SimEnumParameter,
    SimDoubleParameter]
                                       ],
                                  list[tuple[Content, SimultanObject]],
                                  list[tuple[Content, Union[ndarray, DataFrame]]],
                                  list[tuple[Content, FileInfo]],
                                  list[tuple[Content, Any]],
                                  list[tuple[Content, GeometryModel]]:,
                                  ]
    ]:

        parameters = []
        components = []
        arrays = []
        assets = []
        undefined = []
        other = []
        geometry = []

        for content in self.component._taxonomy_map.content:
            val = getattr(self.component, content.property_name)
            if isinstance(val, SimultanObject):
                components.append((content, val))
            elif isinstance(val, (int, float, str)):
                raw_val = self.component.get_raw_attr(content.property_name)
                parameters.append((content, val, raw_val))
            elif isinstance(val, (ndarray, DataFrame)):
                arrays.append((content, val))
            elif isinstance(val, FileInfo):
                assets.append((content, val))
            elif val is None:
                undefined.append((content, val))
            else:
                other.append((content, val))

        return parameters, components, arrays, assets, other, geometry, undefined

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):
        parameters, components, arrays, assets, other, geometry, undefined = self.parameters
        self.ui_param_table(parameters)
        self.ui_component_table(components)
        self.ui_array_table(arrays)
        self.ui_assets_table(assets)
        self.ui_undefined_table(undefined)

    def ui_assets_table(self, assets):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'path', 'label': 'Path', 'field': 'path', 'sortable': True}]

        rows = [{'id': i,
                 'name': asset[0].name,
                 'path': asset[1].name}
                for i, asset in enumerate(assets)]

        asset_table = ui.table(columns=columns,
                               rows=rows,
                               title='Assets',
                               row_key='id').classes('w-full bordered text-sm')

        asset_table.add_slot('body', r'''
            <q-tr :props="props">
                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                    {{ col.value }}
                </q-td>
                <q-td auto-width>
                    <q-btn size="sm" color="blue" round dense
                           @click="$parent.$emit('edit_val', props)"
                           icon="edit" />
                </q-td>
            </q-tr>
        ''')
        asset_table.on('edit_val', self.edit_val)

    def ui_undefined_table(self, other):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'value', 'label': 'Value', 'field': 'value', 'sortable': True}]

        rows = [{'id': i,
                 'name': param[0].name,
                 'value': str(param[1])}
                for i, param in enumerate(other)]

        comp_table = ui.table(columns=columns,
                              rows=rows,
                              title='Undefined Properties',
                              row_key='id').classes('w-full bordered')

        comp_table.add_slot('body', r'''
                            <q-tr :props="props">
                                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                    {{ col.value }}
                                </q-td>
                                <q-td auto-width>
                                    <q-btn size="sm" color="blue" round dense
                                           @click="$parent.$emit('edit_val', props)"
                                           icon="edit" />
                                </q-td>
                            </q-tr>
                        ''')
        comp_table.on('edit_val', self.edit_val)

    def ui_param_table(self, parameters):
        # create int/float/str table
        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'value', 'label': 'Value', 'field': 'value', 'sortable': True},
                   {'name': 'min', 'label': 'Min', 'field': 'min', 'sortable': True},
                   {'name': 'max', 'label': 'Max', 'field': 'max', 'sortable': True},
                   {'name': 'unit', 'label': 'Unit', 'field': 'unit', 'sortable': True},
                   {'name': 'description', 'label': 'Description', 'field': 'description', 'sortable': False}]

        rows = [{'id': i,
                 'name': param[0].name,
                 'value': str(param[2].Value),
                 'min': str(param[2].ValueMin) if hasattr(param[2], 'ValueMin') else '',
                 'max': str(param[2].ValueMax) if hasattr(param[2], 'ValueMax') else '',
                 'unit': param[2].Unit if hasattr(param[2], 'Unit') else '',
                 'description': param[2].Description if hasattr(param[2], 'Description') else ''}
                for i, param in enumerate(parameters)]

        param_table = ui.table(columns=columns,
                               rows=rows,
                               title='Parameters',
                               row_key='id').classes('w-full bordered text-sm')

        param_table.add_slot('body', r'''
            <q-tr :props="props">
                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                    {{ col.value }}
                </q-td>
                <q-td auto-width>
                    <q-btn size="sm" color="blue" round dense
                           @click="$parent.$emit('edit_val', props)"
                           icon="edit" />
                </q-td>
            </q-tr>
        ''')
        param_table.on('edit_val', self.edit_val)

    def ui_component_table(self, components):

        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'component_id', 'label': 'Component ID', 'field': 'component_id', 'sortable': True},
                   {'name': 'component_name', 'label': 'Component Name', 'field': 'component_name', 'sortable': True},
                   {'name': 'component_type', 'label': 'Component Type', 'field': 'component_type', 'sortable': True},
                   {'name': 'type', 'label': 'Type', 'field': 'type', 'sortable': True},
                   ]

        rows = [{'id': str(i),
                 'component_id': str(comp[1].id),
                 'name': comp[0].name,
                 'component_name': comp[1].name,
                 'component_type': comp[1].__class__.__name__,
                 'type': 'Subcomponent' if comp[1]._wrapped_obj in
                                           [y.Component for y in self.component._wrapped_obj.Components]
                 else 'Reference'}
                for i, comp in enumerate(components)]

        comp_table = ui.table(columns=columns,
                              rows=rows,
                              title='Components',
                              row_key='id').classes('w-full bordered text-sm')

        comp_table.add_slot('body', r'''
                    <q-tr :props="props">
                        <q-td v-for="col in props.cols" :key="col.name" :props="props">
                            {{ col.value }}
                        </q-td>
                        <q-td auto-width>
                            <q-btn size="sm" color="blue" round dense
                                   @click="$parent.$emit('detail', props)"
                                   icon="launch" />
                            <q-btn size="sm" color="blue" round dense
                                   @click="$parent.$emit('edit_val', props)"
                                   icon="edit" />
                        </q-td>
                    </q-tr>
                ''')
        comp_table.on('edit_val', self.edit_val)
        comp_table.on('detail', self.show_detail)

    def ui_array_table(self, arrays):
        columns = [{'name': 'id', 'label': 'ID', 'field': 'id', 'sortable': True},
                   {'name': 'name', 'label': 'Name', 'field': 'name', 'sortable': True},
                   {'name': 'type', 'label': 'Type', 'field': 'type', 'sortable': True}]

        rows = [{'id': str(i),
                 'name': str(comp[0].name),
                 'type': 'ND Array' if isinstance(comp[1], ndarray) else 'Table'} for i, comp in enumerate(arrays)]

        comp_table = ui.table(columns=columns,
                              rows=rows,
                              title='Arrays',
                              row_key='id').classes('w-full bordered text-sm')

        comp_table.add_slot('body', r'''
                            <q-tr :props="props">
                                <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                    {{ col.value }}
                                </q-td>
                                <q-td auto-width>
                                    <q-btn size="sm" color="blue" round dense
                                           @click="$parent.$emit('detail', props)"
                                           icon="launch" />
                                    <q-btn size="sm" color="blue" round dense
                                           @click="$parent.$emit('edit_val', props)"
                                           icon="edit" />
                                </q-td>
                            </q-tr>
                        ''')
        comp_table.on('edit_val', self.edit_val)
        comp_table.on('detail', self.show_detail)

    def get_cb_instance(self, e: events.GenericEventArguments):
        content = next((content for content in self.component._taxonomy_map.content
                        if content.name == e.args['row'].get('name')), None)
        instance = getattr(self.component, content.property_name)
        if isinstance(instance, (ndarray, )):
            return self.component.get_raw_attr(content.property_name).ValueSource.Field
        elif isinstance(instance, DataFrame):
            return self.component.get_raw_attr(content.property_name).ValueSource.Table
        else:
            return instance

    def edit_val(self, e: events.GenericEventArguments):

        content = next((content for content in self.component._taxonomy_map.content
                        if content.name == e.args['row'].get('name')), None)
        val = getattr(self.component, content.property_name)
        edit_dialog = ContentEditDialog(component=val,
                                        parent=self,
                                        content=content)
        edit_dialog.create_edit_dialog()

    def refresh(self):
        self.ui_content.refresh()

    def show_detail(self, e: events.GenericEventArguments):
        from ..detail_views import show_detail
        instance = self.get_cb_instance(e)
        show_detail(value=instance)


class MappedClsDetailView(ComponentDetailBaseView):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    @ui.refreshable
    def ui_content(self, *args, **kwargs):
        super().ui_content(*args, **kwargs)
        content_view = ContentView(component=self.component, parent=self)
        content_view.ui_content()


class MappedClsView(TypeView):

    colors = {'item': 'bg-stone-100',
              'cls_color': 'bg-stone-300',
              'selected': 'bg-blue-200'}

    detail_view = MappedClsDetailView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):
        from ..detail_views import show_detail
        # with ui.list().classes('w-full h-full').props('bordered separator'):
        #     with ui.item(on_click=self.show_details).classes('w-full h-full') as self.card:
        #         with ui.item_section().classes('q-ml-auto'):
        #             self.checkbox = ui.checkbox().classes('q-ml-auto')
        #         with ui.item_section().classes('w-full h-full'):
        #             ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name')
        #         with ui.item_section().classes('w-full h-full'):
        #             ui.label(f'{str(self.component.id)}')

        with ui.card().classes(f"{self.colors['item']} w-full h-full") as self.card:
            self.card.on('click', lambda e: show_detail(value=self.component,
                                                        parent=self)
                         )
            with ui.row().classes(f"{self.colors['item']} w-full") as self.row:
                self.row.on('click', lambda e: show_detail(value=self.component,
                                                           parent=self.component)
                            )
                self.checkbox = ui.checkbox(on_change=self.select)
                ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name')
                with ui.row():
                    with ui.row():
                        ui.label('Global ID: ')
                        ui.label(f'{self.component.id.GlobalId.ToString()}')
                    with ui.row():
                        ui.label('Local ID: ')
                        ui.label(f'{self.component.id.LocalId}')

                # with ui.item_section():
                #     ui.button(on_click=self.show_details, icon='launch').classes('q-ml-auto')
            # self.content_view.ui_content()

    def show_details(self, *args, **kwargs):
        self.detail_view(component=self.component,
                         parent=self).ui_content()



import pandas as pd
from typing import Type
from nicegui import Client, app, ui, events
from .type_view import TypeView
from .type_view_manager import TypeViewManager

# from .. import user_manager
# from .detail_views import show_detail
# from ..core.edit_dialog import ContentEditDialog

from SIMULTAN.Data.MultiValues import (SimMultiValueField3D, SimMultiValueField3DParameterSource, SimMultiValueBigTable,
                                       SimMultiValueBigTableHeader, SimMultiValueBigTableParameterSource)

from PySimultan2.multi_values import simultan_multi_value_big_table_to_pandas


class DataFrameDetailView(object):

    def __init__(self, *args, **kwargs) -> None:
        self.component: SimMultiValueBigTable = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.array = None
        self.dim_slider = None
        self.table = None

    def ui_content(self, *args, **kwargs):

        with ui.row().classes('w-full'):
            ui.input(label='Name',
                     value=self.component.Name).classes('w-full').bind_value(self.component,
                                                                                    'Name')
        with ui.row().classes('w-full'):
            with ui.column():
                ui.label('ID:')
                ui.label(f'{str(self.component.Id.GlobalId.ToString())}')
                ui.label(f'{str(self.component.Id.LocalId)}')

        df = simultan_multi_value_big_table_to_pandas(self.component)

        with ui.row().classes('w-full'):
            with ui.column():
                ui.label('Shape:')
                ui.label(f'{df.shape}')

        ui.separator()

        table = ui.table.from_pandas(df).classes('w-full h-full')

        with table.add_slot('top-left'):
            def toggle() -> None:
                table.toggle_fullscreen()
                button.props('icon=fullscreen_exit' if table.is_fullscreen else 'icon=fullscreen')

            button = ui.button('Toggle fullscreen', icon='fullscreen', on_click=toggle).props('flat')


class DataFrameView(TypeView):

    detail_view = DataFrameDetailView

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, value):
        self._component = value

    @ui.refreshable
    def ui_content(self):
        from .detail_views import show_detail

        with ui.card().classes(f"{self.colors['item']} w-full h-full") as self.card:
            self.card.on('click', lambda e: show_detail(value=self.component,
                                                        parent=self))
            with ui.row().classes('bg-stone-100 w-full') as self.row:
                self.row.on('click', lambda e: show_detail(value=self.component,
                                                           parent=self))
                self.checkbox = ui.checkbox(on_change=self.select)
                with ui.row():
                    ui.label('Name:').props('font-weight=bold')
                    ui.input(value=self.component.Name).bind_value(self.component, 'Name')
                    ui.label('ID:').props('font-weight=bold')
                    with ui.row():
                        with ui.row():
                            ui.label(f'{self.component.Id.GlobalId.ToString()}')
                        with ui.row():
                            ui.label(f'{self.component.Id.LocalId}')

    def show_details(self, *args, **kwargs):

        self.detail_view(component=self.component,
                         parent=self).ui_content()

        # TypeView.show_details(self)
        # core.detail_view.clear()
        # with core.detail_view as detail_view:
        #     with ui.card().classes('w-full h-full'):
        #
        #         if kwargs.get('previous', None) is not None:
        #             with ui.row():
        #                 ui.button(on_click=lambda e: kwargs.get('previous').__ui_element__.show_details(previous=self),
        #                           icon='arrow_back').classes('q-mr-md')
        #
        #         with ui.row():
        #             ui.input(label='Name',
        #                      value=self.component.Name).classes('w-full h-full').bind_value(self.component,
        #                                                                                     'Name')
        #         with ui.row():
        #             ui.label('ID:')
        #             ui.label(f'{str(self.component.Id)}')
        #
        #         ui.separator()
        #
        #         table = ui.table.from_pandas(simultan_multi_value_big_table_to_pandas(self.component)
        #                                      ).classes('w-full h-full')
        #
        #         with table.add_slot('top-left'):
        #             def toggle() -> None:
        #                 table.toggle_fullscreen()
        #                 button.props('icon=fullscreen_exit' if table.is_fullscreen else 'icon=fullscreen')
        #
        #             button = ui.button('Toggle fullscreen', icon='fullscreen', on_click=toggle).props('flat')


class DataFrameManager(TypeViewManager):

    cls: pd.DataFrame = pd.DataFrame
    item_view_cls: Type[TypeView] = DataFrameView
    item_view_name = 'Dataframes'

    def button_create_ui_content(self):
        ui.button('Create new DataFrame', on_click=self.create_new_item, icon='add')

    def update_items(self) -> list[SimMultiValueBigTable]:
        if self.data_model is None:
            return []

        return [x for x in self.data_model.value_fields if type(x) == SimMultiValueBigTable]

    @ui.refreshable
    def add_item_to_view(self,
                         item: any,
                         raw_val=None):

        if isinstance(item, SimMultiValueBigTable):
            val_source = item
        elif isinstance(item, pd.DataFrame):
            val_source: SimMultiValueBigTable = raw_val.ValueSource.ValueField

        if self.items_ui_element is None:
            return

        if val_source not in self.items:
            self.items.append(val_source)
        item_view = self.item_views.get(str(val_source.Id), None)

        if item_view is None:
            item_view = self.item_view_cls(component=val_source,
                                           parent=self)
            self.item_views[str(val_source.Id)] = item_view
            with self.items_ui_element:
                item_view.ui_content()
        else:
            if item_view.card.parent_slot.parent.parent_slot.parent is not self.items_ui_element:
                with self.items_ui_element:
                    item_view.ui_content()
        return item_view



from nicegui import ui
import logging
from ..type_view import TypeView
from nicegui import ui, app, events

from ... import user_manager
# from ... import core

from ..component_detail_base_view import ComponentDetailBaseView
from ...core.edit_dialog import ContentEditDialog
from PySimultan2.simultan_object import SimultanObject
from PySimultan2.default_types import ComponentList

from ..parameter_view import ParameterView
from ..mapped_cls.mapped_cls_view import ContentItemView
from ...core.geo_associations import GeoEditDialog

from SIMULTAN.Data import SimId
from System import Guid

logger = logging.getLogger('py_simultan_ui')


class ListItemView(object):

    def __init__(self, *args, **kwargs):
        self.component: SimultanObject = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.item_no = kwargs.get('item_no')
        self.first = kwargs.get('first', False)
        self.last = kwargs.get('last', False)

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):

        from ..detail_views import show_detail

        with ui.item().classes('w-full h-full'):
            with ui.item_section():
                ui.label(f'Item {self.item_no}:')
            if hasattr(self.component, '__ui_element__') and self.component.__ui_element__ is None:
                if self.component.__class__ not in self.view_manager.cls_views:
                    self.view_manager.create_mapped_cls_view_manager(taxonomy=self.component.__class__._taxonomy)
                self.view_manager.cls_views[self.component.__class__]['item_view_manager'].add_item_to_view(
                    self.component
                )
            # with ui.item_section():
            #     ui.label(f'{self.component.name}').on('click', self.component.__ui_element__.show_details)
                # ui.link(self.component.name, self.component.__ui_element__.card)

            with ui.item_section():
                ui.label(f'{self.component.name}')

            with ui.item_section():
                ui.button(on_click=lambda e: show_detail(self.component),
                          icon='launch').classes('q-ml-auto')

            with ui.item_section():
                with ui.row():
                    ui.label(f'{self.component.Id.GlobalId.ToString()}')
                with ui.row():
                    ui.label(f'{self.component.Id.LocalId}')

            with ui.item_section():
                button = ui.button(on_click=self.edit,
                                   icon='edit').classes('q-ml-auto')
                button.item = self.component
                button.item_no = self.item_no
            with ui.item_section():
                with ui.row():
                    if not self.first:
                        button = ui.button(on_click=self.move_up,
                                           icon='keyboard_arrow_up').classes('q-ml-auto')
                        button.item = self.component
                        button.item_no = self.item_no
                    else:
                        ui.label('')
                with ui.row():
                    if not self.last:
                        button = ui.button(on_click=self.move_down,
                                           icon='keyboard_arrow_down').classes('q-ml-auto')
                        button.item = self.component
                        button.item_no = self.item_no
                    else:
                        ui.label('')

            with ui.item_section():
                button = ui.button(on_click=self.remove,
                                   icon='delete').classes('q-ml-auto')
                button.item = self.component
                button.item_no = self.item_no

    def edit(self, event):
        ui.notification('Edit not implemented yet', type='negative')

    def remove(self, event):
        self.parent.component.discard(event.sender.item)
        self.parent.list_content.refresh()

    def move_up(self, event):
        if self.item_no > 0:
            self.parent.component.move_item(event.sender.item, event.sender.item_no - 1)
        self.parent.list_content.refresh()

    def move_down(self, event):
        if self.item_no < len(self.parent.component) - 1:
            self.parent.component.move_item(event.sender.item, event.sender.item_no + 1)
        self.parent.list_content.refresh()


class ListView(object):

    def __init__(self, *args, **kwargs):
        self.component: ComponentList = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.card = None

        self.content_item_views: dict[str: ContentItemView] = {}

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def list_content(self):
        with ui.list().classes('w-full h-full'):

            for i, item in enumerate(self.component):

                if self.content_item_views.get(i, None) is None or self.content_item_views[i].component != item:
                    self.content_item_views[i] = ListItemView(component=item,
                                                              parent=self,
                                                              item_no=i,
                                                              first=i == 0,
                                                              last=i == len(self.component) - 1)
                self.content_item_views[i].ui_content()
                ui.separator()

        try:
            ui.button(f'Add new item to {self.component.name}',
                      on_click=self.parent.add_new_item, icon='add').classes('q-ml-auto')
        except Exception as e:
            logger.error(f'Error adding new item: {e}')

    @ui.refreshable
    def ui_content(self):
        with ui.expansion(icon='format_list_bulleted',
                          text=f'Content ({len(self.component)})',
                          value=True
                          ).classes('w-full h-full').bind_text_from(self,
                                                                    'data',
                                                                    lambda x: f'Content ({len(self.component)})'
                                                                    ) as self.card:
            self.list_content()


class ListDetailView(ComponentDetailBaseView):

    columns = [
        {'name': 'item_id',
         'label': 'Item ID',
         'field': 'item_id',
         'align': 'left',
         'sortable': True},
        {'name': 'name',
         'label': 'Name',
         'field': 'name',
         'sortable': True},
        {'name': 'component_id',
         'label': 'Component ID',
         'field': 'component_id',
         'sortable': True},
        {'name': 'item_type',
         'label': 'Item Type',
         'field': 'item_type',
         'sortable': True},
        {'name': 'type',
         'label': 'Type',
         'field': 'type',
         'sortable': True},
    ]

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    @property
    def items(self) -> list[dict]:
        items = [{}] * len(self.component.data)

        for i, item in enumerate(self.component.data):
            if item._wrapped_obj in self.component.components:
                c_type = 'Subcomponent'
            elif item._wrapped_obj in self.component.ref_components:
                c_type = 'Reference'
            else:
                c_type = 'Undef'

            items[i] = {'item_id': str(i),
                        'name': item.name,
                        'component_id': f'{item.id.GlobalId}_{item.id.LocalId}',
                        'item_type': f'{item.__class__.__name__}',
                        'type': c_type
                        }
        return items

    @ui.refreshable
    def ui_content(self):
        super().ui_content()

        with ui.table(columns=self.columns,
                      rows=self.items,
                      row_key='item_id').classes('w-full bordered') as self.table:

            # self.table.add_slot('header', r'''
            #         <q-tr :props="props">
            #             <q-th auto-width />
            #             <q-th v-for="col in props.cols" :key="col.name" :props="props">
            #                 {{ col.label }}
            #             </q-th>
            #         </q-tr>
            #     ''')

            # Modify the table body slot for delete functionality
            self.table.add_slot('body', r'''
                    <q-tr :props="props">
                        <q-td v-for="col in props.cols" :key="col.name" :props="props">
                            {{ col.value }}
                        </q-td>
                        <q-td auto-width>
                            <q-btn size="sm" color="blue" round dense
                                   @click="$parent.$emit('detail', props)"
                                   icon="launch" />
                            <q-btn size="sm" color="positive" round dense
                                   @click="$parent.$emit('up', props)"
                                   icon="keyboard_arrow_up" />
                            <q-btn size="sm" color="positive" round dense
                                   @click="$parent.$emit('down', props)"
                                   icon="keyboard_arrow_down" />
                            <q-btn size="sm" color="negative" round dense
                                   @click="$parent.$emit('delete', props)"
                                   icon="delete" />
                        </q-td>
                    </q-tr>
                ''')

            # self.table.add_slot('body', r'''
            #     <q-td :props="props">
            #         <q-btn @click="$parent.$emit('delete', props)" icon="delete" flat dense color='green'/>
            #     </q-td>
            # ''')

            self.table.on('delete', self.remove)
            self.table.on('up', self.move_up)
            self.table.on('down', self.move_down)
            self.table.on('detail', self.show_details)

            ui.button(f'Add new item to {self.component.name}',
                      on_click=self.add_new_item, icon='add').classes('q-ml-auto')

        ui.button('Add Item', on_click=self.add_new_item, icon='add')

    def get_instance_row(self, props):
        return self.component[int(props.args['row'].get('item_id'))]

    def remove(self, props):
        instance = self.get_instance_row(props)

        other_refs = [ref.Target for ref in instance._wrapped_obj.ReferencedBy
                      if ref.Target is not self.component._wrapped_obj]

        if other_refs:
            ui.notify(f"Cannot remove {instance.name} because it is referenced by "
                      f'{[(ref.Name, ref.Id) for ref in other_refs]}',
                      type='negative',
                      close_button='OK')
            return

        self.component.discard(instance)
        self.ui_content.refresh()

    def move_up(self, props):
        instance = self.get_instance_row(props)
        item_no = int(props.args['row'].get('item_id'))
        if item_no > 0:
            self.component.move_item(instance, item_no - 1)
        self.ui_content.refresh()

    def move_down(self, props):
        instance = self.get_instance_row(props)
        item_no = int(props.args['row'].get('item_id'))
        if item_no < len(self.parent.component) - 1:
            self.component.move_item(instance, item_no + 1)
        self.ui_content.refresh()

    def show_details(self, props):
        from ..detail_views import show_detail
        instance = self.get_instance_row(props)
        show_detail(value=instance)

    def add_new_item(self, event):

        component = self.component
        edit_dialog = ContentEditDialog(component=None,
                                        parent=self,
                                        content=None,
                                        options=['Component'])

        def save(*args, **kwargs):
            if isinstance(edit_dialog.edit_dialog.value, list):
                component.extend(edit_dialog.edit_dialog.value)
            else:
                component.append(edit_dialog.edit_dialog.value)
            edit_dialog.close()
            self.ui_content.refresh()

        edit_dialog.save = save
        edit_dialog.create_edit_dialog()

    def refresh(self):
        self.ui_content()


class ComponentListView(TypeView):

    detail_view = ListDetailView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def view_manager(self):
        return self.parent.view_manager

    @ui.refreshable
    def ui_content(self):
        from ..detail_views import show_detail

        with ui.card().classes(f"{self.colors['item']} w-full h-full") as self.card:
            self.card.on('click', lambda e: show_detail(self.component))
            with ui.row().classes('bg-stone-100 w-full') as self.row:
                self.row.on('click', lambda e: show_detail(self.component))
                self.checkbox = ui.checkbox()
                ui.input(label='Name', value=self.component.name).bind_value(self.component, 'name')
                ui.label(f'{str(self.component.id)}')

            # self.content_view.ui_content()

    def show_details(self, *args, **kwargs):
        self.detail_view(component=self.component,
                         parent=self).ui_content()

    def add_new_item(self, event):

        component = self.component
        edit_dialog = ContentEditDialog(component=None,
                                        parent=self,
                                        content=None,
                                        options=['Component'])

        def save(self, *args, **kwargs):
            if isinstance(edit_dialog.edit_dialog.value, list):
                component.extend(edit_dialog.edit_dialog.value)
            else:
                component.append(edit_dialog.edit_dialog.value)
            component.__ui_element__.content_view.list_content.refresh()
            edit_dialog.close()

        edit_dialog.save = save
        edit_dialog.create_edit_dialog()





class Class1:
    def __init__(self, *args, **kwargs):
        self.param_a = kwargs.get('param_a', None)
        self.param_b = kwargs.get('param_b', None)
        self.param_c = kwargs.get('param_c', None)

    def add(self):
        self.param_c = self.param_a + self.param_b



from PySimultan2.taxonomy_maps import TaxonomyMap, Content
import PySimultanUI.src.pysimultanui as psui

# example class to be mapped:


class ExampleClass:
    def __init__(self, *args, **kwargs):
        self.a = kwargs.get('a', None)
        self.b = kwargs.get('b', None)
        self.c = None

    def add(self):
        self.c = self.a + self.b

    def subtract(self):
        self.c = self.a - self.b

    def multiply(self):
        self.c = self.a * self.b


# create content and taxonomy map:
contents = {}

contents['a'] = Content(text_or_key='a_in_simultan',
                        property_name='a',
                        type=None,
                        unit=None,
                        documentation='Content a')

contents['b'] = Content(text_or_key='b_in_simultan',
                        property_name='b',
                        type=None,
                        unit=None,
                        documentation='Content b')

contents['c'] = Content(text_or_key='c_in_simultan',
                        property_name='c',
                        type=None,
                        unit=None,
                        documentation='Content c')

example_class_map = TaxonomyMap(taxonomy_name='PySimultanUI',
                                taxonomy_key='PySimultanUI',
                                taxonomy_entry_name='ExampleClass',
                                taxonomy_entry_key='ExampleClass',
                                content=[contents['a'],
                                         contents['b'],
                                         contents['c']
                                         ],
                                )

# get the mapper:
mapper = psui.core.mapper

# register the class to be mapped:
mapper.register(example_class_map.taxonomy_entry_key,
                ExampleClass,
                taxonomy_map=example_class_map)


# register methods:
cls = mapper.get_mapped_class('ExampleClass')

# get the method mapper:
method_mapper = psui.core.method_mapper

method_mapper.register_method(cls=cls,
                              name='subtract',
                              method=cls.subtract,
                              args=[],
                              kwargs={})

method_mapper.register_method(cls=cls,
                              name='add',
                              method=cls.add,
                              args=[],
                              kwargs={})

method_mapper.register_method(cls=cls,
                              name='multiply',
                              method=cls.multiply,
                              args=[],
                              kwargs={})

# run the UI:
psui.run_ui()



from nicegui import ui
from ..type_view_manager import TypeViewManager
from PySimultan2.simultan_object import SimultanObject


class MappedClsManager(TypeViewManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def button_create_ui_content(self):
        ui.button(f'Create new {self.item_view_name}', on_click=self.create_new_item, icon='add')

    def create_new_item(self, event):

        if self.data_model is None:
            ui.notify('No data model selected! Please load a data model first.')
            return

        new_item: SimultanObject = self.cls(data_model=self.data_model,
                                            object_mapper=self.mapper,
                                            )
        self.add_item_to_view(new_item)

    @ui.refreshable
    def ui_content(self):

        with ui.expansion(icon='format_list_bulleted',
                          text=f'{self.item_view_name if self.item_view_name is not None else self.cls._taxonomy} '
                               f'({len(self.items)})'
                          ).classes('w-full h-full bg-stone-100'
                                    ).bind_text_from(self,
                                                     'items',
                                                     lambda x: f'{self.item_view_name} ({len(x)})'
                                                     ) as self.expansion:
            self.methods_content()
            self.ui_expand_content()

    @ui.refreshable
    def ui_expand_content(self):
        self.items_ui_element = ui.row().classes('w-full h-full')
        with self.items_ui_element:
            if len(self.items) == 0:
                with ui.item():
                    self.no_items_label = ui.label('No items to display')

        for item in self.items:
            self.add_item_to_view(item)

        with ui.row().classes('w-full h-full'):
            self.button_create_ui_content()

    @ui.refreshable
    def methods_content(self):

        cls = self.user.mapper.mapped_classes.get(self.cls._taxonomy)
        if cls not in self.user.method_mapper.mapped_methods.keys():
            self.user.method_mapper.resolve_classes()

        mapped_methods = self.user.method_mapper.mapped_methods.get(cls, [])

        with ui.expansion(icon='code', text='Methods').classes('w-full h-full bg-stone-200'):
            with ui.list().classes('w-full h-full'):
                for method in mapped_methods:
                    with ui.item():
                        with ui.item_section():
                            ui.label(method.name)

                        with ui.item_section():
                            ui.button('Run', on_click=method.run).classes('q-ml-auto')



from nicegui import ui
from .scene import Scene

import numpy as np

# Define your vertices and indices
vertices = np.array([
    [0, 0, 0],
    [1, 0, 0],
    [1, 1, 0],
    [0, 1, 0]
], dtype=float).flatten()

indices = np.array([
    0, 1, 2,
    0, 2, 3
], dtype=int)


with Scene() as scene:
    vertices = np.array([
        [0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]
    ], dtype=float)
    indices = np.array([0, 1, 2, 0, 2, 3], dtype=int)
    material_props = {'color': 0xff0000, 'wireframe': True}

    scene.mesh(vertices, indices)

ui.run()




from fastapi import Request
from fastapi.responses import RedirectResponse
from starlette.middleware.base import BaseHTTPMiddleware

from nicegui import app, Client


unrestricted_page_routes = {'/login'}


class AuthMiddleware(BaseHTTPMiddleware):
    """This middleware restricts access to all NiceGUI pages.

    It redirects the user to the login page if they are not authenticated.
    """

    async def dispatch(self, request: Request, call_next):
        if not app.storage.user.get('authenticated', False):
            print(f'Add ID: {id(app)}')
            if request.url.path in Client.page_routes.values() and request.url.path not in unrestricted_page_routes:
                app.storage.user['referrer_path'] = request.url.path  # remember where the user wanted to go
                return RedirectResponse('/login')
        return await call_next(request)



import os
import shutil
import asyncio
from logging import getLogger
from typing import Optional, Type

# from .. import core
from .. import user_manager
from nicegui import Client, app, ui, events
from .type_view import TypeView
from PySimultan2.simultan_object import SimultanObject


logger = getLogger('py_simultan_ui')


class TypeViewManager(object):

    cls: Optional[Type[SimultanObject]] = None
    item_view_cls: type[TypeView] = TypeView
    item_view_name = 'Item'

    def __init__(self, *args, **kwargs):

        self._mapped_cls = None
        self._data_model = None
        self._items = []
        self._item_views: dict[any, any] = {}

        self.items_ui_element = kwargs.get('items_ui_element', None)
        self.no_items_label = None

        self.data_model = kwargs.get('data_model', None)
        self.expansion: Optional[ui.expansion] = kwargs.get('expansion', None)

    @property
    def project_manager(self):
        return self.user.project_manager

    @property
    def mapper(self):
        return self.user.mapper

    @property
    def mapped_cls(self):
        if self._mapped_cls is None:
            self._mapped_cls = self.mapper.mapped_classes.get(self.taxonomy, None)
        return self._mapped_cls

    @property
    def user(self):
        return user_manager[app.storage.user['username']]

    @property
    def taxonomy(self):
        if self.cls is not None:
            return self.cls._taxonomy if hasattr(self.cls, '_taxonomy') else self.cls

    @property
    def items(self) -> list[any]:
        if self._items is None:
            self.items = self.update_items()
        return self._items

    @items.setter
    def items(self, value: list[any]):

        old_value = self._items
        for removed_item in [x for x in old_value if x not in value]:
            self.remove_item_from_view(removed_item)
        self._items = value
        for item in self.items:
            self.add_item_to_view(item)

        self.update_no_items_label()

    @property
    def item_views(self):
        return self._item_views

    @item_views.setter
    def item_views(self, value: dict[any, any]):
        self._item_views = value

    @property
    def data_model(self):
        return self._data_model

    @property
    def selected_items(self):
        return [x for x in self.items if x.__ui_element__.selected]

    @data_model.setter
    def data_model(self, value):
        logger.debug(f'View Manager setting data model to {value}')
        # self._item_views = {}
        self._data_model = value
        self._items = self.update_items()
        self.ui_content.refresh()

    @ui.refreshable
    def ui_content(self):

        logger.info(f'Creating UI content for TVM {self.taxonomy}')

        with ui.row().classes('w-full h-full') as self.expansion:

        # with ui.expansion(icon='format_list_bulleted',
        #                   text=f'{self.item_view_name if self.item_view_name is not None else self.cls._taxonomy} '
        #                        f'({len(self.items)})'
        #                   ).classes('w-full h-full').bind_text_from(self,
        #                                                             'items',
        #                                                             lambda x: f'{self.item_view_name} ({len(x)})'
        #                                                             ) as self.expansion:
            self.ui_expand_content()

    @ui.refreshable
    def ui_expand_content(self):

        logger.info(f'Creating UI expand content for TVM {self.taxonomy}')

        self.items_ui_element = ui.row().classes('w-full h-full')
        with self.items_ui_element:
            if len(self.items) == 0:
                with ui.item():
                    self.no_items_label = ui.label('No items to display')

        for item in self.items:
            self.add_item_to_view(item)

        with ui.row().classes('w-full h-full'):
            self.button_create_ui_content()

    def button_create_ui_content(self):
        ui.button('Upload new Asset', on_click=self.create_new_item, icon='add')

    def update_items(self) -> list[any]:



        logger.info(f'Updating items for TVM {self.taxonomy}')

        # return self.cls.cls_instances

        if self.cls is None:
            return []
        mapped_cls = self.mapped_cls

        if mapped_cls is None:
            logger.warning(f'mapped_cls for TVM {self.taxonomy} is None!')
            return []
        if hasattr(mapped_cls, 'cls_instances'):
            logger.info(f'Found {len(mapped_cls.cls_instances)} instances for TVM {self.taxonomy}')
            return mapped_cls.cls_instances
        return []

    def create_new_item(self, event):

        logger.info(f'Updating items for TVM {self.taxonomy}')

        if self.data_model is None:
            ui.notify('No data model selected! Please load a data model first.')
            return

        new_item = self.cls()
        self.add_item_to_view(new_item)

    def update_expansion_text(self):
        if self.expansion is not None:
            self.expansion.text = f'{self.item_view_name if self.item_view_name is not None else self.cls._taxonomy} ' \
                                  f'({len(self.items)})'

    def add_item_to_view(self, item: any):

        logger.info(f'Adding item to TVM {self.taxonomy} view: {item}')

        if self.items_ui_element is None:
            return

        if item not in self.items:
            self.items.append(item)
        item_view: TypeView = self.item_views.get(item, None)

        if item_view is None:
            item_view: TypeView = self.item_view_cls(component=item,
                                                     parent=self)
            self.item_views[item] = item_view
            item.__ui_element__ = self.item_views[item]
            with self.items_ui_element:
                item_view.ui_content()
        else:
            if item_view.card.parent_slot.parent.parent_slot.parent is not self.items_ui_element:
                with self.items_ui_element:
                    item_view.ui_content()

        self.update_expansion_text()
        return item_view

    def remove_item_from_view(self, item: any):

        logger.info(f'Removing item from TVM {self.taxonomy} view: {item}')

        try:
            item_view = self.item_views.get(item, None)
            if item_view is not None:
                self.items_ui_element.remove(item_view.card)
                self.item_views.pop(item)
                item.__ui_element__ = None
        except Exception as e:
            pass

    def update_no_items_label(self):
        if self.items.__len__() == 0 and self.no_items_label is None and self.items_ui_element is not None:
            with self.items_ui_element:
                self.no_items_label = ui.label('No items to display')
        elif self.items.__len__() > 0 and self.no_items_label is not None and self.items_ui_element is not None:
            try:
                self.items_ui_element.remove(self.no_items_label)
                self.no_items_label = None
            except Exception as e:
                pass

    def new_instance_created(self, instance: any):
        self.add_item_to_view(instance)

    def update_items_views(self):

        logger.info(f'Updating item views for TVM {self.taxonomy}')

        self.item_views = {}
        self._items = self.update_items()
        print(f'Updating items: {len(self.items)}')
        self.ui_expand_content.refresh()



from contextlib import contextmanager
from nicegui import app, ui, context
from .. import user_manager

from ..__about__ import __version__
from . import home

from .. import core
from ..core.mappers import MapperDropdown


@contextmanager
def frame(navtitle: str):

    """Custom page frame to share the same styling and behavior across all pages"""
    # ui.colors(primary='#da291c', secondary='#941c13', accent='#941c13', positive='#53B689')
    with ui.header().classes('justify-between h-16 shadow-md') as header:

        user = user_manager.users[app.storage.user['username']]

        # ui.image('web_ui/src/static_files/A1_Digital_identifier_pos_RGB.png').classes('w-32')

        with ui.column().classes('flex items-center'):
            with ui.row():
                ui.label('PySimultan').classes('font-bold text-xl')
                ui.label(f'Version {__version__}').classes('text-sm text-gray-500 justify-end items-center')

        with ui.column().classes('flex items-center'):
            with ui.row():

                ui.label(f'Hello {user.name}')
                ui.space()

        with ui.column():
            tool_select = MapperDropdown(user=user_manager.users[app.storage.user['username']]).classes('font-bold text-sm')
            user_manager.users[app.storage.user['username']].tool_select = tool_select

        with ui.column().classes('justify=end items=center'):
            with ui.row():

                with ui.tabs().classes('h-10') as tabs:
                    # ui.button('Mapped Classes', icon='checklist_rtl', on_click=lambda: left_drawer.toggle())
                    ui.tab('Home', icon='home').classes('h-10')
                    ui.tab('Project', icon='edit').classes('h-10')
                    ui.tab('Logs', icon='format_list_bulleted').classes('h-10')

                # mapper_select_dialog = core.mappers.MapperSelectDialog()
                user_settings_dialog = core.user.UserSettingsDialog()
                add_toolbox_dialog = core.mappers.AddPackageDialog()


                with ui.button(icon='menu'):
                    with ui.menu() as menu:

                        # def on_click_mappers():
                        #     setattr(mapper_select_dialog, 'user', user_manager.users[app.storage.user['username']])
                        #     mapper_select_dialog.open()
                        #
                        # ui.menu_item('Toolbox',
                        #              on_click=on_click_mappers
                        #              )

                        def on_click_user_settings():
                            setattr(user_settings_dialog, 'user', user_manager.users[app.storage.user['username']])
                            user_settings_dialog.open()

                        ui.menu_item('User Settings',
                                     on_click=on_click_user_settings
                                     )

                        def add_toolbox():
                            setattr(add_toolbox_dialog, 'user', user_manager.users[app.storage.user['username']])
                            add_toolbox_dialog.open()

                        ui.menu_item('Add Toolbox',
                                     on_click=add_toolbox
                                     )

                        ui.menu_item('Reload Project',
                                     on_click=lambda: user_manager.users[app.storage.user['username']].
                                     project_manager.open_data_model()
                                     )
                        ui.menu_item('Save Project',
                                     on_click=lambda: (user_manager.users[app.storage.user['username']].data_model.save(),
                                                       ui.notify('Project saved!'))
                                     )
                        ui.menu_item('Close Project',
                                     on_click=lambda: user_manager.users[app.storage.user['username']].
                                     project_manager.close_data_model()
                                     )
                        ui.menu_item('Logout',
                                     on_click=lambda: (user_manager.users[app.storage.user['username']].logout(),
                                                       app.storage.user.clear(),
                                                       ui.navigate.to('/login')))

                        ui.separator()
                        ui.menu_item('Close', menu.close)

                # ui.button(on_click=lambda: (app.storage.user.clear(),
                #                             ui.open('/login')),
                #           icon='logout').props('outline round').classes('text-black')
                #


                # with ui.button(icon='menu'):
                #     with ui.menu() as menu:
                #         ui.menu_item('Logout',
                #                      icon='logout',
                #                      on_click=lambda: (app.storage.user.clear(), ui.open('/login')))

                # ui.space()
                # if app.storage.user.get('username', False):
                #     ui.label(f'{app.storage.user["username"]}').classes('mr-auto text-2xl')

    # with ui.left_drawer(bordered=True,
    #                     top_corner=False,
    #                     elevated=True,
    #                     fixed=True,
    #                     value=False).classes('bg-blue-100') as left_drawer:
    #     user.navigation_drawer = left_drawer

    # with ui.splitter() as splitter:
    #     with splitter.after:
    #         with ui.right_drawer(bordered=True,
    #                              top_corner=False,
    #                              elevated=True,
    #                              fixed=True,
    #                              value=False).classes('w-full h-full') as detail_view:
    #             core.detail_view = detail_view

    with ui.tab_panels(tabs, value='Home').classes('w-full h-full'):
        with ui.tab_panel('Project').classes('w-full h-full') as project_full_tab:
            with ui.splitter(value=60, limits=(10, 90)).classes('w-full h-full') as splitter:
                with splitter.before as project_tab:
                    context.client.content.classes('h-[92vh]')
                    user.project_tab = project_tab
                with splitter.after:
                    user.detail_view = ui.row().classes('w-full h-full')

        user.home_tab = ui.tab_panel('Home').classes('w-full h-full')

        with ui.tab_panel('Logs').classes('w-full h-full'):
            with ui.card().classes('w-full h-full') as logs_tab:
                user.log_tab = logs_tab

    with user.home_tab:
        home.content()



import numpy as np
import pandas as pd
from pandas import DataFrame
# import pandas as pd
import pdb

from typing import Any, Optional, TYPE_CHECKING
from nicegui import ui, app
from logging import getLogger
from PySimultan2.simultan_object import SimultanObject
from PySimultan2.geometry import GeometryModel
from PySimultan2.default_types import ComponentDictionary, ComponentList
from PySimultan2.files import FileInfo
from PySimultan2.data_model import DataModel

from SIMULTAN.Data.MultiValues import (SimMultiValueField3D, SimMultiValueField3DParameterSource, SimMultiValueBigTable,
                                       SimMultiValueBigTableHeader, SimMultiValueBigTableParameterSource)

from .. import user_manager

from .pandas_df_view import DataFrameDetailView
from .numpy_view import NDArrayDetailView
from .geometry_view.geometry_view import GeometryDetailView
from .component_list_view.component_list_view import ListDetailView
from .component_dict_view.component_dict_view import DictDetailView
from .asset_view.asset_view import AssetDetailView
from .mapped_cls.mapped_cls_view import MappedClsDetailView

if TYPE_CHECKING:
    from ..core.user import User


logger = getLogger('py_simultan_ui')


def show_next_detail(*args, **kwargs):
    user = user_manager[app.storage.user['username']]
    show_detail(value=user.detail_history.move_next(), *args, **kwargs)


def show_previous_detail(*args, **kwargs):
    user = user_manager[app.storage.user['username']]
    show_detail(value=user.detail_history.move_previous(), *args, **kwargs)


class DetailView(object):

    def __init__(self, *args, **kwargs) -> None:
        """
        Class which represents a detail view.
        :param args:
        :param kwargs:
        """

        self.component: Any = kwargs.get('component')
        self.parent = kwargs.get('parent')

    def ui_content(self, *args, **kwargs):
        """
        Method to create the ui content for the detail view of self.component.
        :param args:
        :param kwargs:
        :return:
        """
        return

    @property
    def user(self) -> Optional['User']:
        return user_manager.users[app.storage.user['username']]

    @property
    def data_model(self) -> Optional[DataModel]:
        return self.user.data_model


def show_detail(value, *args, **kwargs):
    user = user_manager[app.storage.user['username']]
    if value is None:
        user.detail_view.clear()
        user.detail_history.clear()
        return

    user.detail_history.add_item(value)
    # current_detail = user.detail_history[-1] if user.detail_history else None
    logger.debug(f'Showing details for {value}')
    user.detail_view.clear()
    with (user.detail_view):
        with ui.card().classes('w-full'):
            with ui.row():
                previous_detail = user.detail_history.get_previous_detail()
                if previous_detail is not None:
                    ui.button(on_click=show_previous_detail,
                              icon='arrow_back').classes('q-mr-md')
                ui.space()
                next_detail = user.detail_history.get_next_detail()
                if next_detail is not None:
                    ui.button(on_click=show_next_detail,
                              icon='arrow_forward').classes('q-mr-md')

            user.current_detail = value

            if hasattr(value, '_taxonomy_map') and value._taxonomy_map.taxonomy_entry_key is not None:
                detail_view_cls = user.mapping.view_manager.views.get(value._taxonomy_map.taxonomy_entry_key, None)
            else:
                detail_view_cls = None

            if detail_view_cls is None:
                if isinstance(value, SimMultiValueBigTable):
                    detail_view = DataFrameDetailView(component=value, **kwargs)
                elif isinstance(value, SimMultiValueField3D):
                    detail_view = NDArrayDetailView(component=value, **kwargs)
                elif isinstance(value, GeometryModel):
                    detail_view = GeometryDetailView(component=value, **kwargs)
                elif isinstance(value, ComponentList):
                    detail_view = ListDetailView(component=value, **kwargs)
                elif isinstance(value, ComponentDictionary):
                    detail_view = DictDetailView(component=value, **kwargs)
                elif isinstance(value, SimultanObject):
                    detail_view = MappedClsDetailView(component=value, **kwargs)
                elif isinstance(value, FileInfo):
                    detail_view = AssetDetailView(component=value, **kwargs)
            else:
                detail_view = detail_view_cls(component=value, **kwargs)
            detail_view.ui_content()



from .. import user_manager
from nicegui import app
from .. import core
from ..core.project_manager import ProjectManager

from ..core import method_mapper


def content() -> None:
    # project_manager = ProjectManager()
    # get user project manager
    project_manager = user_manager[app.storage.user['username']].project_manager
    # core.project_manager = project_manager
    # # core.project_manager.geometry_file_manager = project_manager.geometry_file_manager
    project_manager.ui_content()
    # core.asset_manager.ui_content()



from typing import List, Optional
from PySimultan2.object_mapper import PythonMapper
from nicegui import ui
import os

from PySimultanUI.src.pysimultanui import user_manager, Content, TaxonomyMap, run_ui, DetailView

from PySimultan2 import DataModel


project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)

new_data_model = DataModel.create_new_project(project_path=os.path.join(project_dir, 'mapper_test.simultan'),
                                              user_name='admin',
                                              password='admin')


def create_first_mapping():
    mapping = user_manager.mapper_manager.create_mapping(name='Toolbox 1')
    mapper = mapping.mapper
    method_mapper = mapping.method_mapper
    view_manager = mapping.view_manager

    contents = {}

    class Class1:
        def __init__(self, *args, **kwargs):
            self.param_a = kwargs.get('param_a', None)
            self.param_b = kwargs.get('param_b', None)
            self.param_c = kwargs.get('param_c', None)

        def add(self):
            self.param_c = self.param_a + self.param_b

    contents['param_a'] = Content(text_or_key='param_a',
                                  property_name='param_a',
                                  type=None,
                                  unit=None,
                                  documentation='param_a')

    contents['param_b'] = Content(text_or_key='param_b',
                                  property_name='param_b',
                                  type=None,
                                  unit=None,
                                  documentation='param_b')

    contents['param_c'] = Content(text_or_key='param_c',
                                  property_name='param_c',
                                  type=None,
                                  unit=None,
                                  documentation='param_c')

    cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                          taxonomy_key='PySimultan',
                          taxonomy_entry_name='Class1',
                          taxonomy_entry_key='Class1',
                          content=[contents['param_a'],
                                   contents['param_b'],
                                   contents['param_c'],
                                   ],
                          )

    mapper.register(cls_map.taxonomy_entry_key, Class1, taxonomy_map=cls_map)
    mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)

    method_mapper.register_method(cls=mapped_cls,
                                  name='add',
                                  method=Class1.add)

    # create detail view
    class Class1DetailView(DetailView):

        def ui_content(self, *args, **kwargs):
            ui.label('This is a detail view for Class1')
            ui.label('param_a')
            ui.input(value=self.component.param_a).bind_value(self.component, 'param_a')
            ui.input(value=self.component.param_b).bind_value(self.component, 'param_b')

    view_manager.views[cls_map.taxonomy_entry_key] = Class1DetailView

    return mapper


def create_second_mapping():
    mapping = user_manager.mapper_manager.create_mapping(name='Toolbox 2')
    mapper = mapping.mapper
    method_mapper = mapping.method_mapper
    view_manager = mapping.view_manager

    contents = {}

    class Class2:
        def __init__(self, *args, **kwargs):
            self.param_a = kwargs.get('param_a', None)
            self.param_b = kwargs.get('param_b', None)
            self.param_c = kwargs.get('param_c', None)
            self.param_d = kwargs.get('param_d', None)

        def multiply(self):
            self.param_d = self.param_a * self.param_b

    class Class3:
        def __init__(self, *args, **kwargs):
            self.thermal_conductivity = kwargs.get('thermal_conductivity', 15)
            self.thermal_resistance = kwargs.get('thermal_resistance', 30)

        def test_method(self):
            print(f'This is a test method {self}')


    contents['param_a'] = Content(text_or_key='param_a',
                                  property_name='param_a',
                                  type=None,
                                  unit=None,
                                  documentation='param_a')

    contents['param_b'] = Content(text_or_key='param_b',
                                  property_name='param_b',
                                  type=None,
                                  unit=None,
                                  documentation='param_b')

    contents['param_c'] = Content(text_or_key='param_c',
                                  property_name='param_c',
                                  type=None,
                                  unit=None,
                                  documentation='param_c')

    contents['param_d'] = Content(text_or_key='param_d',
                                  property_name='param_d',
                                  type=None,
                                  unit=None,
                                  documentation='param_d')

    contents['thermal_conductivity'] = Content(text_or_key='thermal_conductivity',
                                               property_name='thermal_conductivity',
                                               type=None,
                                               unit=None,
                                               documentation='thermal_conductivity')

    contents['thermal_resistance'] = Content(text_or_key='thermal_resistance',
                                             property_name='thermal_resistance',
                                             type=None,
                                             unit=None,
                                             documentation='thermal_resistance')

    cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                          taxonomy_key='PySimultan',
                          taxonomy_entry_name='Class2',
                          taxonomy_entry_key='Class2',
                          content=[contents['param_a'],
                                   contents['param_b'],
                                   contents['param_c'],
                                   contents['param_d'],
                                   ],
                          )

    mapper.register(cls_map.taxonomy_entry_key, Class2, taxonomy_map=cls_map)
    mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)

    method_mapper.register_method(cls=mapped_cls,
                                  name='multiply',
                                  method=Class2.multiply)

    cls_map2 = TaxonomyMap(taxonomy_name='PySimultan',
                           taxonomy_key='PySimultan',
                           taxonomy_entry_name='Class3',
                           taxonomy_entry_key='Class3',
                           content=[contents['thermal_conductivity'],
                                    contents['thermal_resistance'],
                                    ],
                           )

    mapper.register(cls_map2.taxonomy_entry_key, Class3, taxonomy_map=cls_map2)
    mapped_cls2 = mapper.get_mapped_class(cls_map2.taxonomy_entry_key)

    method_mapper.register_method(cls=mapped_cls2,
                                  name='test_method',
                                  method=Class3.test_method)

    def test_fcn(*args, **kwargs):
        print('Test function')

    method_mapper.register_method(method=test_fcn,
                                  name='Test function',
                                  add_user_to_kwargs=True,
                                  add_data_model_to_kwargs=True,
                                  cls=None, args=[],
                                  kwargs={})

    return mapper


def create_components(data_model: DataModel,
                      mapper: 'PythonMapper',
                      mapper2: 'PythonMapper' = None):
    cls = mapper.get_mapped_class('Class1')
    cls(name='class 1 mapper 1',
        param_a=1,
        param_b=2)

    cls2 = mapper2.get_mapped_class('Class2')
    cls2(param_a=3, param_b=4, name='Test Mapper2 class2')

    cls3 = mapper2.get_mapped_class('Class3')
    cls3(name='Test Mapper2 class3', thermal_conductivity=20, thermal_resistance=40)

    data_model.save()
    data_model.cleanup()


mapper1 = create_first_mapping()
mapper2 = create_second_mapping()
create_components(new_data_model, mapper1, mapper2)

run_ui()
print('Done!')



from nicegui import Client, app, ui, events


class ConfirmationDialog(ui.dialog):

    def __init__(self, question: str, *options: str, additional_content_fcn=None, **kwargs):
        super().__init__(value=True)
        self.props("persistent")
        with self, ui.card():
            ui.label(question).classes("text-lg")

            if additional_content_fcn:
                additional_content_fcn()

            with ui.row():
                for c_option in options:
                    ui.button(c_option, on_click=lambda e, option=c_option: self.submit(option))



from nicegui.elements.scene_object3d import Object3D
import numpy as np


class Mesh(Object3D):

    def __init__(self, vertices: np.ndarray,
                 indices: np.ndarray,
                 **kwargs):

        # super().__init__('sphere',
        #                  vertices.tolist(),
        #                  indices.tolist()
        #                  )
        super().__init__('mesh',
                         vertices.tolist(),
                         indices.tolist()
                         )


class Wire(Object3D):

    def __init__(self,
                 vertices: np.ndarray,
                 **kwargs):

        super().__init__('wire',
                         vertices.tolist()
                         )


class Spline(Object3D):
    """
    A spline object using CatmullRomCurve3
    """

    def __init__(self,
                 vertices: np.ndarray,
                 **kwargs):

        super().__init__('spline',
                         vertices.tolist()
                         )



import os
import shutil
import tempfile
from typing import Optional
from nicegui import ui, events
from . import AssetView
from PySimultan2.files import FileInfo, create_asset_from_file

from ..type_view_manager import TypeViewManager
from ... import core


class AssetManager(TypeViewManager):

    columns = [
        {'name': 'id',
         'label': 'Key',
         'field': 'id',
         'align': 'left',
         'sortable': True},
        {'name': 'name',
         'label': 'Name',
         'field': 'name',
         'sortable': True},
        {'name': 'size',
         'label': 'File Size',
         'field': 'size',
         'align': 'left',
         'sortable': True},
        {'name': 'last_modified',
         'label': 'Last modified',
         'field': 'last_modified',
         'align': 'left',
         'sortable': True}
    ]

    item_view_name = 'Assets'
    item_view_cls = AssetView
    cls = FileInfo

    @property
    def items(self) -> list[FileInfo]:
        return self.update_items()

    @items.setter
    def items(self, value):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.table: Optional[ui.table] = None

    def update_items(self) -> list[any]:
        if self.data_model is None:
            return []
        assets = self.data_model.project_data_manager.AssetManager.Resources
        return [FileInfo(resource_entry=asset) for asset in assets]

    def button_create_ui_content(self):
        ui.button('Upload new Asset', on_click=self.create_new_item)

    def create_new_item(self, event):
        if self.data_model is None:
            ui.notify('No data model selected! Please select a data model first.')
            return

        with ui.dialog() as dialog, ui.card():
            ui.upload(label='Upload asset',
                      on_upload=self.upload_project).on(
                'finish', lambda: ui.notify('Finish!')
            ).classes('max-w-full')
            ui.button('Cancel', on_click=lambda e: dialog.close()).classes('mt-4')

        dialog.open()

    def upload_project(self,
                       e: events.UploadEventArguments,
                       *args,
                       **kwargs):

        with tempfile.TemporaryDirectory() as tmpdirname:
            temp_file_path = os.path.join(tmpdirname, e.name)
            with open(temp_file_path, 'wb') as f:
                shutil.copyfileobj(e.content, f)
        # local_path = f'/tmp/{e.name}'
        # shutil.copyfileobj(e.content, open(local_path, 'wb'))
            ui.notify(f'Project {e.name} uploaded!')
            new_fi = FileInfo(file_path=temp_file_path)
            new_asset = create_asset_from_file(new_fi,
                                               data_model=self.data_model,
                                               tag=None)
        self.add_item_to_view(FileInfo(resource_entry=new_asset,
                                       data_model=self.data_model))

    def add_item_to_view(self, asset: FileInfo):
        if self.table is not None:
            self.table.add_rows({'id': f'{asset.resource_entry.Key}',
                                 'name': asset.name,
                                 'size': f'{asset.file_size / 1024:.3f} KB' if asset.file_size/1024 < 1024
                                 else f'{asset.file_size / 1024 / 1024:.3f} MB',
                                 'last_modified': asset.last_modified
                                 })
            self.table.run_method('scrollTo', len(self.table.rows) - 1)

    @ui.refreshable
    def ui_content(self):

        # self.update_items()

        rows = [{'id': f'{asset.resource_entry.Key}',
                 'name': asset.name,
                 'size': f'{asset.file_size / 1024:.3f} KB' if asset.file_size/1024 < 1024
                 else f'{asset.file_size / 1024 / 1024:.3f} MB',
                 'last_modified': asset.last_modified
                 }
                for asset in self.items]

        with ui.table(columns=self.columns,
                      rows=rows,
                      title='Assets',
                      row_key='id').classes('w-full bordered') as self.table:

            self.table.add_slot('body', r'''
                                <q-tr :props="props">
                                    <q-td v-for="col in props.cols" :key="col.name" :props="props">
                                        {{ col.value }}
                                    </q-td>
                                    <q-td auto-width>
                                        <q-btn size="sm" color="blue" round dense
                                                   @click="$parent.$emit('show_detail', props)"
                                                   icon="launch" />
                                        <q-btn size="sm" color="blue" round dense
                                               @click="$parent.$emit('download', props)"
                                               icon="download" />
                                    </q-td>
                                </q-tr>
                            ''')

            self.table.on('download', self.download)
            self.table.on('show_detail', self.show_details)

        self.button_create_ui_content()

    def download(self, e: events.GenericEventArguments):
        asset = next(asset for asset in self.items if asset.resource_entry.Key == int(e.args['row']['id']))
        ui.download(f'assets/{asset.name}')

    def update_items_views(self):

        self.item_views = {}
        self._items = self.update_items()
        print(f'Updating items: {len(self.items)}')
        self.ui_content.refresh()

    def get_instance_row(self, props):
        return next((asset for asset in self.items if asset._resource_entry.Key == int(props.args['row']['id'])), None)

    def show_details(self, props):
        from ..detail_views import show_detail
        instance = self.get_instance_row(props)
        show_detail(value=instance)



from typing import Optional, Type
from nicegui import ui, app

from .. import user_manager
from ..import core
from ..views.type_view_manager import TypeViewManager
from ..views.view_manager import ViewManager


class Navigation(object):

    def __init__(self, *args, **kwargs):
        pass
        # self._view_manager = None
        # self.data_model = kwargs.get('data_model', core.data_model)
        # self.navigation_drawer = core.navigation_drawer
        # self.view_manager: Optional[Type[ViewManager]] = kwargs.get('view_manager', core.view_manager)
        # self.geometry_manager: Optional[Type[TypeViewManager]] = kwargs.get('geometry_manager', core.geometry_manager)
        # self.asset_manager: Type[TypeViewManager] = kwargs.get('asset_manager')

    @property
    def data_model(self):
        return user_manager[app.storage.user['username']].data_model

    @property
    def navigation_drawer(self):
        return user_manager[app.storage.user['username']].navigation_drawer

    @property
    def view_manager(self):
        return user_manager[app.storage.user['username']].view_manager

    @property
    def geometry_manager(self):
        return user_manager[app.storage.user['username']].geometry_manager

    @property
    def asset_manager(self):
        return user_manager[app.storage.user['username']].asset_manager

    @property
    def obj_mapper(self) -> core.PythonMapper:
        return self.view_manager.mapper

    @ui.refreshable
    def ui_content(self):
        with self.navigation_drawer:
            with ui.expansion(icon='format_list_bulleted',
                              text=f'Mapped Classes ({len(self.view_manager.cls_views)})').classes('w-full'):

                for cls, cls_view_dict in self.view_manager.cls_views.items():
                    item_view_manager = cls_view_dict.get('item_view_manager', None)
                    if item_view_manager is None:
                        continue
                    else:
                        with ui.row():
                            def add_link(cls, cls_view_dict):
                                taxonomy = cls._taxonomy if hasattr(cls, '_taxonomy') else cls.__name__
                                cls_instances = cls.cls_instances if hasattr(cls, 'cls_instances') else []

                                ui.link(f'{taxonomy} ({len(cls_instances)})',
                                        cls_view_dict['item_view_manager'].expansion
                                        ).bind_text_from(cls,
                                                         'cls_instances'
                                                         , lambda x: f'{cls._taxonomy} ({len(x)})')

                            add_link(cls, cls_view_dict)

            if self.geometry_manager.expansion is not None:
                ui.link('Geometry models', self.geometry_manager.expansion)

            if self.asset_manager.expansion is not None:
                ui.link('Assets', self.asset_manager.expansion)



import random
from logging import getLogger
from PySimultan2.geometry.geometry_base import (GeometryModel, SimultanVertex, SimultanEdge, SimultanEdgeLoop,
                                                SimultanFace, SimultanVolume, SimultanLayer)

import FreeCAD
import Part as FCPart

from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from PySimultan2.data_model import DataModel
    from PySimultan2.object_mapper import PythonMapper

logger = getLogger('PySimultan-FreeCAD-Tools',
                   )
logger.setLevel('DEBUG')


def import_freecad(doc: FreeCAD.Document,
                   geo_model: GeometryModel,
                   data_model: Optional['DataModel'] = None,
                   object_mapper: Optional['PythonMapper'] = None,
                   scale: float=1.0) -> tuple[int, int, int, int, int]:

    vertex_lookup: dict[FCPart.Vertex: SimultanVertex] = {}
    vertex_pos_lookup: dict[tuple[float, float, float]: SimultanVertex] = {}
    edge_lookup: dict[FCPart.Edge: SimultanEdge] = {}
    edge_v_lookup: dict[tuple[SimultanVertex, SimultanVertex]: SimultanEdge] = {}
    wire_lookup: dict[FCPart.Wire: SimultanEdgeLoop] = {}
    wire_e_lookup: dict[tuple[SimultanEdge]: SimultanEdgeLoop] = {}
    face_lookup: dict[FCPart.Face: SimultanFace] = {}
    face_w_lookup: dict[tuple[SimultanEdgeLoop]: SimultanFace] = {}
    volume_lookup: dict[FCPart.Volume: SimultanVolume] = {}

    layer = SimultanLayer(name='Layer 1',
                          geometry_model=geo_model,
                          data_model=data_model if data_model is not None else geo_model._data_model,
                          object_mapper=object_mapper if object_mapper is not None else geo_model._object_mapper)

    def vertex_from_freecad(vertex: FCPart.Vertex) -> SimultanVertex:
        if vertex in vertex_lookup:
            return vertex_lookup[vertex]
        elif (vertex.X * scale, vertex.Y * scale, vertex.Z * scale) in vertex_pos_lookup:
            return vertex_pos_lookup[(vertex.X * scale, vertex.Y * scale, vertex.Z * scale)]

        new_vertex = SimultanVertex(x=vertex.X * scale,
                                    y=vertex.Y * scale,
                                    z=vertex.Z * scale,
                                    layer=layer)
        vertex_pos_lookup[(vertex.X * scale,
                           vertex.Y * scale,
                           vertex.Z * scale)] = new_vertex
        vertex_lookup[vertex] = new_vertex
        return new_vertex

    def edge_from_freecad(edge: FCPart.Edge) -> SimultanEdge:
        if edge in edge_lookup:
            return edge_lookup[edge]
        elif (edge.Vertexes[0], edge.Vertexes[1]) in edge_v_lookup:
            return edge_v_lookup[(edge.Vertexes[0], edge.Vertexes[1])]

        new_edge = SimultanEdge(vertices=[vertex_from_freecad(edge.Vertexes[0]),
                                          vertex_from_freecad(edge.Vertexes[1])],
                                layer=layer)
        edge_lookup[edge] = new_edge
        edge_v_lookup[(edge.Vertexes[0], edge.Vertexes[1])] = new_edge

        return new_edge

    def wire_from_freecad(wire: FCPart.Wire) -> SimultanEdgeLoop:
        print(f'Wire: {wire.Edges}')
        if wire in wire_lookup:
            return wire_lookup[wire]

        edges = [edge_from_freecad(edge) for edge in wire.OrderedEdges]

        if tuple(edges) in wire_e_lookup:
            return wire_e_lookup[tuple(edges)]

        new_wire = SimultanEdgeLoop(edges=edges,
                                    layer=layer)
        wire_lookup[wire] = new_wire
        wire_e_lookup[tuple(edges)] = new_wire

        return new_wire

    def face_from_freecad(face: FCPart.Face) -> SimultanFace:
        print(f'Face: {face.Wires}')

        if face in face_lookup:
            return face_lookup[face]
        wires = [wire_from_freecad(wire) for wire in face.Wires]

        if tuple(wires) in face_w_lookup:
            return face_w_lookup[tuple(wires)]

        boundary_edge_loop = wire_from_freecad(face.OuterWire)
        holes = [wire_from_freecad(x) for x in [x for x in face.Wires if not x.isEqual(face.OuterWire)]]

        if holes:
            logger.debug(f'Face has holes: {holes}')

        new_face = SimultanFace(edge_loop=boundary_edge_loop,
                                holes=holes,
                                layer=layer)
        new_face._wrapped_object.Color.set_Color(new_face._wrapped_object.Color.
                                                 Color.FromRgb(random.randint(0, 255),
                                                               random.randint(0, 255),
                                                               random.randint(0, 255)
                                                               )
                                                 )
        face_lookup[face] = new_face
        face_w_lookup[tuple(wires)] = new_face
        return new_face

    def volume_from_freecad(volume: FCPart.Solid) -> SimultanVolume:
        print(f'Volume: {volume.Faces}')

        if volume in volume_lookup:
            return volume_lookup[volume]

        faces = [face_from_freecad(face) for face in volume.Faces]
        new_volume = SimultanVolume(faces=faces,
                                    layer=layer)

        volume_lookup[volume] = new_volume
        return new_volume

    for feature in doc.Objects:
        try:
            if isinstance(feature.Shape, FCPart.Vertex):
                vertex = vertex_from_freecad(feature)
            elif isinstance(feature.Shape, FCPart.Edge):
                edge = edge_from_freecad(feature)
            elif isinstance(feature.Shape, FCPart.Wire):
                wire = wire_from_freecad(feature.Shape)
            elif isinstance(feature.Shape, FCPart.Face):
                face = face_from_freecad(feature.Shape)
            elif isinstance(feature.Shape, FCPart.Solid):
                volume = volume_from_freecad(feature.Shape)
        except Exception as e:
            print(f'Error processing feature: {e}')
            continue

    return (vertex_lookup.__len__(), edge_lookup.__len__(),
            wire_lookup.__len__(), face_lookup.__len__(), volume_lookup.__len__())



from typing import Union
from .user import UserManager

from PySimultan2.object_mapper import PythonMapper
from PySimultan2.data_model import DataModel
from PySimultan2 import config as PySimultan2_config
from .method_mapper import method_mapper, mapped_method
from .logging import logger

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..app.home import ProjectManager
    from .navigation import Navigation


mapper = PythonMapper()
method_mapper.mapper = mapper
mapper.method_mapper = method_mapper

PySimultan2_config.logger.setLevel('DEBUG')
PySimultan2_config.set_default_mapper(mapper)

print('mapper:', mapper)



import os
import asyncio
from nicegui import ui, app
from .mappers import MapperManager
from typing import Union, Optional

from .logging import logger

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .method_mapper import MethodMapper
    from PySimultan2.data_model import DataModel
    from PySimultan2.object_mapper import PythonMapper

initial_user_name = os.environ.get('INITIAL_USER_NAME', 'admin')
initial_user_email = os.environ.get('INITIAL_USER_EMAIL', 'example@test.de')
initial_user_password = os.environ.get('INITIAL_USER_PASSWORD', 'admin')

logger.info(f'User module imported')
class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class DetailHistory:

    max_history_entries = 10

    def __init__(self):
        self.detail_history = []
        self.detail_history_index = 0
        self.current_detail = None

    def add_item(self, item):

        if len(self.detail_history) > 0:
            if item is self.detail_history[self.detail_history_index]:
                return

        if len(self.detail_history) > 0 and self.detail_history_index < (len(self.detail_history) - 1):
            self.detail_history = self.detail_history[:self.detail_history_index]
            self.detail_history_index = len(self.detail_history) - 1

        self.detail_history.append(item)
        if len(self.detail_history) > self.max_history_entries:
            self.detail_history.pop(0)
        self.detail_history_index = len(self.detail_history) - 1
        self.current_detail = item

    def get_current_detail(self):
        return self.current_detail

    def get_previous_detail(self):
        if self.detail_history_index > 0:
            return self.detail_history[self.detail_history_index - 1]

    def get_next_detail(self):
        if self.detail_history_index < (len(self.detail_history) - 1):
            return self.detail_history[self.detail_history_index + 1]

    def __len__(self):
        return len(self.detail_history)

    def move_next(self):
        if self.detail_history_index < len(self.detail_history) - 1:
            self.detail_history_index += 1
            return self.detail_history[self.detail_history_index]

    def move_previous(self):
        if self.detail_history_index > 0:
            self.detail_history_index -= 1
            return self.detail_history[self.detail_history_index]

    def clear(self):
        self.detail_history = []
        self.detail_history_index = 0
        self.current_detail = None


class User:

    def __init__(self, *args, **kwargs):

        from ..core import method_mapper, mapper

        self._project_manager = None
        self._grid_view = None
        self._view_manager = None
        self._asset_manager = None
        self._geometry_manager = None
        self._array_manager = None
        self._navigation = None
        self._data_model = None
        self._ui_log = None
        self._log_tab = None
        self._selected_mapper = None
        self._load_undefined = True

        self.user_manager = kwargs.get('user_manager', None)
        self.detail_history = DetailHistory()

        self.name = kwargs.get('name', None)
        self.email = kwargs.get('email', None)
        self.password = kwargs.get('password', None)

        # self.mapper: Optional[PythonMapper] = kwargs.get('mapper', mapper)
        # self.method_mapper: Optional[MethodMapper] = kwargs.get('method_mapper', method_mapper)
        self.data_model: Union[DataModel, None] = kwargs.get('data_model', None)

        self.navigation_drawer = kwargs.get('navigation_drawer', None)
        self.project_tab = kwargs.get('project_tab', None)
        self.log_tab = kwargs.get('log_tab', None)
        self.detail_view = kwargs.get('detail_view', None)
        self.home_tab = kwargs.get('home_tab', None)
        self.tool_select = kwargs.get('tool_select', None)

    @property
    def mapper_manager(self):
        return self.user_manager.mapper_manager

    @property
    def mapping(self):
        return self.mapper_manager.get_mapping(user=self,
                                               name=self.selected_mapper,
                                               load_undefined=self.load_undefined)

    @property
    def mapper(self):
        if self.mapping is None:
            return None

        mapper = self.mapping.mapper
        return mapper

    @property
    def method_mapper(self):
        return self.mapping.method_mapper

    @mapper.setter
    def mapper(self, value):
        self.mapping.mapper = value

    @property
    def load_undefined(self):
        return self._load_undefined

    @load_undefined.setter
    def load_undefined(self, value):
        if value == self._load_undefined:
            return
        self._load_undefined = value
        self.mapper.load_undefined = value
        self.mapper.clear()
        self.project_manager.open_data_model()
        self.project_manager.refresh_all_items()

    @property
    def mapper_options(self) -> list[str]:
        options = list(self.mapper_manager.available_mappings.keys())
        options.sort()
        return options

    @property
    def selected_mapper(self) -> Optional[str]:
        if self._selected_mapper is None:
            self._selected_mapper = self.mapper_options[0] if len(
                self.mapper_manager.available_mappings) > 0 else None
        return self._selected_mapper

    @selected_mapper.setter
    def selected_mapper(self, value):
        if value == self._selected_mapper:
            return

        old_mapper = self.mapper
        old_mapper.clear()

        self._selected_mapper = value
        self.mapper = self.mapper.copy()
        self.mapper.load_undefined = self.load_undefined
        self.method_mapper.ui_content.refresh()
        self.project_manager.open_data_model()

    @property
    def available_mappings(self):
        return self.mapper_manager.available_mappings

    @property
    def data_model(self):
        return self._data_model

    @data_model.setter
    def data_model(self, value):
        self._data_model = value
        if self.detail_history is not None:
            self.detail_history = DetailHistory()

    @property
    def grid_view(self):
        if self._grid_view is None:
            from ..views.grid_view import GridView
            self._grid_view = GridView()
        return self._grid_view

    @property
    def project_manager(self):
        if self._project_manager is None:
            from .project_manager import ProjectManager
            self._project_manager = ProjectManager(user_manager=self.user_manager,
                                                   mapper=self.mapper)
        return self._project_manager

    @property
    def view_manager(self):
        # if self._view_manager is None:
        #     from ..views.view_manager import ViewManager
        #     self._view_manager = ViewManager(mapper=self.mapper,
        #                                      parent=self.project_manager)
        return self._view_manager

    @property
    def asset_manager(self):
        if self._asset_manager is None:
            from ..views.asset_view import AssetManager
            self._asset_manager = AssetManager()
        return self._asset_manager

    @property
    def array_manager(self):
        if self._array_manager is None:
            from ..views.array_view import ArrayManager
            self._array_manager = ArrayManager()
        return self._array_manager

    @property
    def geometry_manager(self):
        if self._geometry_manager is None:
            from ..views.geometry_view import GeometryManager
            self._geometry_manager = GeometryManager(data_model=self.data_model)
        return self._geometry_manager

    @property
    def navigation(self):
        # if self._navigation is None:
        #     from .navigation import Navigation
        #     self._navigation = Navigation()
        return self._navigation

    @property
    def log_tab(self):
        return self._log_tab

    @log_tab.setter
    def log_tab(self, value):
        self._log_tab = value

        if self._log_tab is not None and self._ui_log is None:
            self.add_ui_log()

    @property
    def ui_log(self):
        if self._ui_log is None and self.log_tab is not None:
            self.add_ui_log()
        return self._ui_log

    @property
    def logger(self):
        if self.ui_log is None:
            return None
        return self.ui_log.logger

    def logout(self):
        self.data_model = None

    def add_ui_log(self):
        if self.log_tab is None:
            return
        from .ui_log import UILog
        with self.log_tab:
            self._ui_log = UILog()
            self._ui_log.ui_content()

    def select_mapper(self, *args, **kwargs):

        pass

        # with ui.dialog() as dialog, ui.card():
        #     ui.label('Select mapper:')
        #
        #     select = ui.select(options=list(self.available_mappings.keys()),
        #                        value=self.selected_mapper)
        #
        #     with ui.row():
        #         ui.button('OK', on_click=lambda: dialog.submit({'ok': True,
        #                                                         'mapper_name': select.value}
        #                                                        )
        #                   )
        #         ui.button('Cancel', on_click=lambda: dialog.submit({'ok': False}))

        # mapper_select_dialog = MapperSelectDialog(user=self)
        # mapper_select_dialog.open()

        # await asyncio.sleep(0.01)
        #
        # if not result.get('ok', False):
        #     return
        #
        # self.selected_mapper = result.get('mapper_name', self.selected_mapper)
        # self.project_manager.refresh_all_items()
        #
        # ui.notify(f'Selected mapper: {self.selected_mapper}')


class Admin(User):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UserManager(metaclass=Singleton):

    def __init__(self):
        logger.debug('UserManager created')
        self.mapper_manager = MapperManager()
        self.users = {}
        self.create_initial_users()

    @property
    def current_user(self):
        return self.users[app.storage.user['username']]

    @property
    def current_data_model(self):
        return self.current_user.data_model

    def create_initial_users(self):
        admin = Admin(name=initial_user_name,
                      email=initial_user_email,
                      password=initial_user_password,
                      user_manager=self)

        self.users[initial_user_name] = admin

        user_1 = User(name=initial_user_name + '_2',
                      email=initial_user_email + '_2',
                      password=initial_user_password + '_2',
                      user_manager=self)

        self.users[initial_user_name + '_2'] = user_1

    def authenticate(self, username, password):
        if username in self.users and self.users[username].password == password:
            return True
        return False

    def __getitem__(self, key):
        return self.users[key]


class UserSettingsDialog(ui.dialog):

    def __init__(self, user=None) -> None:
        super().__init__()
        self._user = None

        with self, ui.card().classes('w-full h-full'):
            ui.label('User settings:')
            with ui.checkbox('Load undefined components',
                             value=True) as self.load_undefined:
                ui.tooltip('Load components that are not defined in the mapping')

            with ui.row().classes('justify-center'):
                ui.button('OK', on_click=self.ok)
                ui.button('Cancel', on_click=self.cancel)

        self.user = user

    @property
    def user(self) -> 'User':
        return self._user

    @user.setter
    def user(self, value: 'User') -> None:
        self._user = value
        if self._user is not None:
            self.load_undefined.value = self._user.load_undefined
        else:
            self.load_undefined.value = True

    def cancel(self, *args, **kwargs):
        self.close()

    def ok(self, *args, **kwargs):
        print('User settings saved')
        self.user.load_undefined = self.load_undefined.value
        ui.notify('User settings saved')
        self.close()

    def open(self, *args, **kwargs):
        super().open(*args, **kwargs)



import os
from nicegui import ui, app
from nicegui.elements.scene import Scene as NGScene
from PySimultanUI.src.pysimultanui.geometry_3d.scene import ExtendedScene
from PySimultanUI.src.pysimultanui.geometry_3d.three_views import (display_vertices,
                                                                   display_edges,
                                                                   display_wire,
                                                                   display_face,
                                                                   display_solid)

import numpy as np

import FreeCAD
import Part as FCPart

# Define your vertices and indices
vertices = np.array([
    [0, 0, 0],
    [1, 0, 0],
    [1, 1, 0],
    [0, 1, 0]
], dtype=float).flatten()

indices = np.array([
    0, 1, 2,
    0, 2, 3
], dtype=int)


@ui.page('/3d_test')
def index_page1() -> None:
    with ui.card():
        ui.label('3D Test')
        with ExtendedScene().classes('w-full h-full') as scene:
            vertices = np.array([
                [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1],
                [0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0],
            ], dtype=float)
            indices = np.array([[0, 1, 2,],
                                [0, 2, 3],
                                [4, 5, 6],
                                [4, 6, 7],
                                [0, 4, 5],
                                [0, 5, 1],
                                [1, 5, 6],
                                [1, 6, 2],
                                [2, 6, 7],
                                [2, 7, 3],
                                [3, 7, 4],
                                [3, 4, 0]
                                ], dtype=int)
            # material_props = {'color': 0xff0000, 'wireframe': True}
            scene.curve([-4, 0, 0], [-4, -1, 0], [-3, -1, 0], [-3, 0, 0]).material('#008800')
            # scene.sphere().material('#4488ff')
            scene.mesh(vertices, indices).material('#4488ff')

            scene.wire(np.array([[-4, 0, 0], [-4, -1, 0], [-3, -1, 0], [-3, 0, 0]])).material('#ff0000')


storage_secret = os.environ.get('STORAGE_SECRET', 'my_secret6849')

# set_storage_secret(storage_secret)
ui.run(storage_secret=storage_secret,
       title='Py Simultan',
       dark=False,
       reload=True,
       port=8080,
       uvicorn_logging_level='info',
       )



import sys
import traceback
from typing import Any, Optional, List, Union
from .scene import ExtendedScene
import numpy as np

import logging

from SIMULTAN.Data.Geometry import (Layer, Vertex, Edge, PEdge, Face, Volume, EdgeLoop, OperationPermission,
                                             GeometryModelOperationPermissions, GeometryOperationPermissions,
                                             LayerOperationPermissions, GeometryModelData, GeometricOrientation,
                                             BaseEdgeContainer)

from PySimultan2.geometry.geometry_base import (SimultanVolume, SimultanFace, SimultanEdge, SimultanVertex,
                                                SimultanLayer, SimultanEdgeLoop, GeometryModel)



logger = logging.getLogger('PySimultanUI')


def rgb_to_hex(rgb: List[int]) -> str:
    if rgb is None:
        return '#000000'
    return f'#{rgb[0] << 16 | rgb[1] << 8 | rgb[2]:06x}'


def display_simgeo(simgeo: Union[Layer, Vertex, Edge, PEdge, Face, Volume, EdgeLoop],
                   scene: ExtendedScene,
                   obj_id: str = '',
                   color: Optional[list[int]] = None):
    try:

        with scene.group() as group:
            rgb_color = rgb_to_hex([simgeo._wrapped_object.Color.Color.R,
                                    simgeo._wrapped_object.Color.Color.G,
                                    simgeo._wrapped_object.Color.Color.B])

            if isinstance(simgeo, SimultanVertex):
                scene.point_cloud([simgeo.position],
                                  colors=[[simgeo._wrapped_object.Color.Color.R,
                                           simgeo._wrapped_object.Color.Color.G,
                                           simgeo._wrapped_object.Color.Color.B]],
                                  point_size=0.1).with_name(simgeo.id)
            elif isinstance(simgeo, SimultanEdge):
                scene.line(simgeo.vertex_0.position,
                           simgeo.vertex_1.position).material(rgb_color).with_name(simgeo.id)
            elif isinstance(simgeo, SimultanFace):
                tris = simgeo.triangulate()
                scene.mesh(tris[0], tris[1]).material(rgb_color).with_name(simgeo.id)
            elif isinstance(simgeo, SimultanVolume):
                for face in simgeo.faces:
                    tris = face.triangulate()
                    scene.mesh(tris[0], tris[1]).material(rgb_color).with_name(simgeo.id)

            return group

    except Exception as e:
        logger.error(f'Error displaying simgeo: {e}\n'
                     f'{traceback.format_exception(*sys.exc_info())}')



from typing import List, Optional
import os
import numpy as np
import pandas as pd
from PySimultanUI.src.pysimultanui.main_ui import run_ui
from PySimultan2 import config
from PySimultanUI.src.pysimultanui.core import mapper, method_mapper, mapped_method
from PySimultan2.taxonomy_maps import TaxonomyMap, Content
from PySimultan2.files import FileInfo
from PySimultan2 import DataModel
from PySimultan2.geometry.geometry_base import (GeometryModel)
from PySimultan2.geometry.utils import create_cube
# from py_simultan_ui.src.core.method_mapper import method_mapper, mapped_method

project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)

new_data_model = DataModel.create_new_project(project_path=os.path.join(project_dir, 'steady_state.simultan'),
                                              user_name='admin',
                                              password='admin')


def create_geometry_model(name='new_geometry_test',
                          data_model: DataModel = new_data_model):
    return GeometryModel(name=name,
                         data_model=data_model)


def create_classes() -> dict[str, type]:

    class TypeTestClass(object):

        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name', 'Type test class')
            self.int_value = kwargs.get('int_value', 1459)
            self.float_value = kwargs.get('float_value', 259.0)
            self.str_value = kwargs.get('str_value', 'string value')
            self.bool_value = kwargs.get('bool_value', True)
            self.list_value = kwargs.get('list_value', [])
            self.dict_value = kwargs.get('dict_value', {'a': 1, 'b': 2})
            self.none_value = kwargs.get('none_value', None)
            self.obj_value = kwargs.get('obj_value')
            self.numpy_value = kwargs.get('numpy_value', np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                                                                  dtype=np.int32).reshape(4, 3))
            self.pandas_value = kwargs.get('pandas_value', pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]}))
            self.file_value = kwargs.get('file_value', None)

    class Source(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.temperature = kwargs.get('temperature')
            self.heat_flux = kwargs.get('heat_flux')

    class TemperatureBoundaryCondition(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.temperature = kwargs.get('temperature', 293.0)

    class HeatFluxBoundaryCondition(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.heat_flux = kwargs.get('heat_flux')

    class Material(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.thermal_conductivity = kwargs.get('thermal_conductivity', 0.15)
            self.density = kwargs.get('density', 1000)
            self.specific_heat = kwargs.get('specific_heat', 800)

    class WallLayer(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.thickness = kwargs.get('thickness', 0.10)
            self.material: Optional[Material] = kwargs.get('material', None)

        @property
        def r_value(self):
            return self.thickness / self.material.thermal_conductivity

    class Construction(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.layers: List[WallLayer] = kwargs.get('layers', [])
            self.r_value = kwargs.get('r_value', None)

        @mapped_method(name='calculate_r_value')
        def calculate_r_value(self):
            self.r_value = sum([layer.r_value for layer in self.layers])
            return self.r_value

    class Wall(object):
        h_c_lookup = {0.1: 5,
                      0.13: 7.69,
                      0.17: 5.88}

        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.r_si = kwargs.get('r_si')
            self.r_se = kwargs.get('r_se')
            self.construction: Construction = kwargs.get('construction')
            self.u_value = kwargs.get('u_value', None)

            self.boundary_condition = kwargs.get('boundary_condition', None)

        @property
        def area(self):
            return sum((x.area for x in self.associated_geometry))

        @property
        def h_c(self):
            """
            return the convective heat transfer coefficient h_c in W/m2K
            :return:
            """
            return self.h_c_lookup[self.r_si]

        def calculate_u_value(self) -> float:
            self.u_value = 1 / (self.r_si + self.r_se + self.construction.calculate_r_value())
            return self.u_value

    class Zone(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.walls: List[Wall] = kwargs.get('walls', [])
            self.volume = kwargs.get('volume')
            self.area = kwargs.get('area')
            self.u_value = kwargs.get('u_value', None)

            self.sources: List[Source] = kwargs.get('sources', [])
            self.steady_state_temperature = kwargs.get('steady_state_temperature', None)

        def calculate_area(self):
            self.area = self.associated_geometry[0].area_brutto_netto.Item1
            return self.area

        def calculate_steady_state_temperature(self):
            phi_ti = []
            for wall in self.walls:
                phi_i = wall.area * wall.calculate_u_value() * wall.r_si * wall.h_c
                phi_ti.append((phi_i * wall.boundary_condition.temperature, phi_i))

            phi_hc_id = sum(x.heat_flux for x in self.sources)

            t_steady = (sum(x[0] for x in phi_ti) + phi_hc_id) / sum(x[1] for x in phi_ti)
            self.steady_state_temperature = t_steady

            return t_steady

    class Volume(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.geometry_file = kwargs.get('geometry_file')

    class Mesh(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name')
            self.mesh_file = kwargs.get('mesh_file')

    class Assembly(object):
        def __init__(self, *args, **kwargs):
            self.name = kwargs.get('name', 'Unnamed Assembly')
            self.volumes = kwargs.get('volumes', [])
            self.mesh = kwargs.get('mesh', None)
            self.features = kwargs.get('features', [])

    class ReferenceFace(object):
        def __init__(self, *args, **kwargs):
            self.construction: Construction = kwargs.get('construction')
            self.assembly: Assembly = kwargs.get('assembly')
            self.mesh_size = 15

    cls_dict = {'Material': Material,
                'WallLayer': WallLayer,
                'Wall': Wall,
                'Zone': Zone,
                'TemperatureBoundaryCondition': TemperatureBoundaryCondition,
                'HeatFluxBoundaryCondition': HeatFluxBoundaryCondition,
                'Source': Source,
                'Construction': Construction,
                'TypeTestClass': TypeTestClass,
                'Assembly': Assembly,
                'ReferenceFace': ReferenceFace,
                'Volume': Volume,
                'Mesh': Mesh}

    return cls_dict


def create_mapped_classes(classes: dict[str, type]) -> dict[str, type]:

    def create_contents() -> dict[str, Content]:
        contents = {}

        contents['thermal_conductivity'] = Content(text_or_key='thermal_conductivity',
                                                   property_name='thermal_conductivity',
                                                   type=None,
                                                   unit='W/mK',
                                                   documentation='thermal_conductivity in W/mK')

        contents['density'] = Content(text_or_key='density',
                                      property_name='density',
                                      type=None,
                                      unit='kg/m3',
                                      documentation='density in kg/m3')

        contents['specific_heat'] = Content(text_or_key='specific_heat',
                                            property_name='specific_heat',
                                            type=None,
                                            unit='J/kgK',
                                            documentation='specific_heat in J/kgK')

        contents['thickness'] = Content(text_or_key='thickness',
                                        property_name='thickness',
                                        type=None,
                                        unit='m',
                                        documentation='thickness in m')

        contents['material'] = Content(text_or_key='material',
                                       property_name='material',
                                       type=None,
                                       unit=None,
                                       documentation='material')

        contents['construction'] = Content(text_or_key='construction',
                                           property_name='construction',
                                           type=None,
                                           unit=None,
                                           documentation='construction')

        contents['r_si'] = Content(text_or_key='r_si',
                                   property_name='r_si',
                                   type=None,
                                   unit='m2K/W',
                                   documentation='r_si in m2K/W')

        contents['r_se'] = Content(text_or_key='r_se',
                                   property_name='r_se',
                                   type=None,
                                   unit=None,
                                   documentation='r_se in m2K/W')

        contents['layers'] = Content(text_or_key='layers',
                                     property_name='layers',
                                     type=None,
                                     unit=None,
                                     documentation='layers of the wall',
                                     component_policy='subcomponent')

        contents['u_value'] = Content(text_or_key='u_value',
                                      property_name='u_value',
                                      type=None,
                                      unit='W/m2K',
                                      documentation='u_value of the wall in W/m2K')

        contents['r_value'] = Content(text_or_key='r_value',
                                      property_name='r_value',
                                      type=None,
                                      unit='m2K/W',
                                      documentation='r_value of the wall in W/m2K')

        contents['boundary_condition'] = Content(text_or_key='boundary_condition',
                                                 property_name='boundary_condition',
                                                 type=None,
                                                 unit=None,
                                                 documentation='boundary_condition')

        contents['walls'] = Content(text_or_key='walls',
                                    property_name='walls',
                                    type=None,
                                    unit=None,
                                    documentation='walls of the zone')

        contents['volume'] = Content(text_or_key='volume',
                                     property_name='volume',
                                     type=None,
                                     unit=None,
                                     documentation='volume of the zone in m3')

        contents['area'] = Content(text_or_key='area',
                                   property_name='area',
                                   type=None,
                                   unit='m2',
                                   documentation='area of the zone in m2')

        contents['sources'] = Content(text_or_key='sources',
                                      property_name='sources',
                                      type=None,
                                      unit=None,
                                      documentation='sources of the zone')

        contents['steady_state_temperature'] = Content(text_or_key='steady_state_temperature',
                                                       property_name='steady_state_temperature',
                                                       type=None,
                                                       unit='K',
                                                       documentation='steady_state_temperature of the zone')

        contents['temperature'] = Content(text_or_key='temperature',
                                          property_name='temperature',
                                          type=None,
                                          unit='K',
                                          documentation='temperature in K')

        contents['heat_flux'] = Content(text_or_key='heat_flux',
                                        property_name='heat_flux',
                                        type=None,
                                        unit='W',
                                        documentation='heat_flux in W')

        contents['int_value'] = Content(text_or_key='int_value',
                                        property_name='int_value',
                                        type=int,
                                        unit=None,
                                        documentation='int_value')

        contents['float_value'] = Content(text_or_key='float_value',
                                          property_name='float_value',
                                          type=float,
                                          unit=None,
                                          documentation='float_value')

        contents['str_value'] = Content(text_or_key='str_value',
                                        property_name='str_value',
                                        type=str,
                                        unit=None,
                                        documentation='str_value')

        contents['bool_value'] = Content(text_or_key='bool_value',
                                         property_name='bool_value',
                                         type=bool,
                                         unit=None,
                                         documentation='bool_value')

        contents['list_value'] = Content(text_or_key='list_value',
                                         property_name='list_value',
                                         type=list,
                                         unit=None,
                                         documentation='list_value')

        contents['dict_value'] = Content(text_or_key='dict_value',
                                         property_name='dict_value',
                                         type=dict,
                                         unit=None,
                                         documentation='dict_value')

        contents['none_value'] = Content(text_or_key='none_value',
                                         property_name='none_value',
                                         type=None,
                                         unit=None,
                                         documentation='none_value')

        contents['obj_value'] = Content(text_or_key='obj_value',
                                        property_name='obj_value',
                                        type=None,
                                        unit=None,
                                        documentation='obj_value')

        contents['numpy_value'] = Content(text_or_key='numpy_value',
                                          property_name='numpy_value',
                                          type=None,
                                          unit=None,
                                          documentation='numpy_value')

        contents['pandas_value'] = Content(text_or_key='pandas_value',
                                           property_name='pandas_value',
                                           type=None,
                                           unit=None,
                                           documentation='pandas_value')

        contents['assembly'] = Content(text_or_key='assembly',
                                       property_name='assembly',
                                       type=None,
                                       unit=None,
                                       documentation='assembly')

        contents['volumes'] = Content(text_or_key='volumes',
                                      property_name='volumes',
                                      type=None,
                                      unit=None,
                                      documentation='volumes')

        contents['features'] = Content(text_or_key='features',
                                       property_name='features',
                                       type=None,
                                       unit=None,
                                       documentation='features')

        contents['mesh'] = Content(text_or_key='mesh',
                                   property_name='mesh',
                                   type=None,
                                   unit=None,
                                   documentation='mesh')

        contents['geometry_file'] = Content(text_or_key='geometry_file',
                                            property_name='geometry_file',
                                            type=None,
                                            unit=None,
                                            documentation='geometry_file')

        contents['mesh_file'] = Content(text_or_key='mesh_file',
                                        property_name='mesh_file',
                                        type=None,
                                        unit=None,
                                        documentation='mesh_file')

        contents['mesh_size'] = Content(text_or_key='mesh_size',
                                        property_name='mesh_size',
                                        type=float,
                                        unit='cm',
                                        documentation='mesh_size')

        return contents

    def create_mapped_type_test_class(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='TypeTestClass',
                              taxonomy_entry_key='TypeTestClass',
                              content=[contents['int_value'],
                                       contents['float_value'],
                                       contents['str_value'],
                                       contents['bool_value'],
                                       contents['list_value'],
                                       contents['dict_value'],
                                       contents['none_value'],
                                       contents['obj_value'],
                                       contents['numpy_value'],
                                       contents['pandas_value']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_mapped_material(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Material',
                              taxonomy_entry_key='Material',
                              content=[contents['thermal_conductivity'],
                                       contents['density'],
                                       contents['specific_heat']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_layer_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Layer',
                              taxonomy_entry_key='Layer',
                              content=[contents['thickness'], contents['material']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_construction_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Construction',
                              taxonomy_entry_key='Construction',
                              content=[contents['layers'], contents['r_value']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_wall_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Wall',
                              taxonomy_entry_key='Wall',
                              content=[contents['r_si'],
                                       contents['r_se'],
                                       contents['construction'],
                                       contents['u_value'],
                                       contents['boundary_condition']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_zone_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Zone',
                              taxonomy_entry_key='Zone',
                              content=[contents['walls'],
                                       contents['volume'],
                                       contents['area'],
                                       contents['u_value'],
                                       contents['sources'],
                                       contents['steady_state_temperature']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_temperature_bc_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='TemperatureBoundaryCondition',
                              taxonomy_entry_key='TemperatureBoundaryCondition',
                              content=[contents['temperature']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_heat_flux_bc_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='HeatFluxBoundaryCondition',
                              taxonomy_entry_key='HeatFluxBoundaryCondition',
                              content=[contents['heat_flux']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_source_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Source',
                              taxonomy_entry_key='Source',
                              content=[contents['temperature'],
                                       contents['heat_flux']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_reference_face_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='ReferenceFace',
                              taxonomy_entry_key='ReferenceFace',
                              content=[contents['construction'],
                                       contents['assembly'],
                                       contents['mesh_size']
                                       ],
                              )
        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_assembly_cls(cls, contents: dict[str, Content]):

        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Assembly',
                              taxonomy_entry_key='Assembly',
                              content=[contents['volumes'],
                                       contents['mesh'],
                                       contents['features']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_volume_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Volume',
                              taxonomy_entry_key='Volume',
                              content=[contents['geometry_file']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    def create_mesh_cls(cls, contents: dict[str, Content]):
        cls_map = TaxonomyMap(taxonomy_name='PySimultan',
                              taxonomy_key='PySimultan',
                              taxonomy_entry_name='Mesh',
                              taxonomy_entry_key='Mesh',
                              content=[contents['mesh_file']],
                              )

        mapper.register(cls_map.taxonomy_entry_key, cls, taxonomy_map=cls_map)
        mapped_cls = mapper.get_mapped_class(cls_map.taxonomy_entry_key)
        return mapped_cls

    tax_contents = create_contents()

    mapped_type_test_class = create_mapped_type_test_class(classes['TypeTestClass'], tax_contents)
    mapped_material_cls = create_mapped_material(classes['Material'], tax_contents)
    mapped_layer_cls = create_layer_cls(classes['WallLayer'], tax_contents)
    mapped_construction_cls = create_construction_cls(classes['Construction'], tax_contents)
    mapped_wall_cls = create_wall_cls(classes['Wall'], tax_contents)
    mapped_zone_cls = create_zone_cls(classes['Zone'], tax_contents)
    mapped_temperature_bc_cls = create_temperature_bc_cls(classes['TemperatureBoundaryCondition'], tax_contents)
    mapped_heat_flux_bc_cls = create_heat_flux_bc_cls(classes['HeatFluxBoundaryCondition'], tax_contents)
    mapped_source_cls = create_source_cls(classes['Source'], tax_contents)
    mapped_reference_face_cls = create_reference_face_cls(classes['ReferenceFace'], tax_contents)
    mapped_assembly_cls = create_assembly_cls(classes['Assembly'], tax_contents)
    mapped_volume_cls = create_volume_cls(classes['Volume'], tax_contents)
    mapped_mesh_cls = create_mesh_cls(classes['Mesh'], tax_contents)

    mapped_cls_dict = {'Material': mapped_material_cls,
                       'Layer': mapped_layer_cls,
                       'Construction': mapped_construction_cls,
                       'Wall': mapped_wall_cls,
                       'Zone': mapped_zone_cls,
                       'TemperatureBoundaryCondition': mapped_temperature_bc_cls,
                       'HeatFluxBoundaryCondition': mapped_heat_flux_bc_cls,
                       'Source': mapped_source_cls,
                       'TypeTestClass': mapped_type_test_class,
                       'Assembly': mapped_assembly_cls,
                       'ReferenceFace': mapped_reference_face_cls,
                       'Volume': mapped_volume_cls,
                       'Mesh': mapped_mesh_cls}

    return mapped_cls_dict


def create_new_steady_state_zone(mapped_classes,
                                 data_model: Optional[DataModel] = None):
    geo_model = create_geometry_model(name='new_geometry_test',
                                      data_model=data_model)
    cube = create_cube(data_model, geo_model, scale=10)
    TypeTestClass = mapped_classes['TypeTestClass']
    Material = mapped_classes['Material']
    WallLayer = mapped_classes['Layer']
    Construction = mapped_classes['Construction']
    Wall = mapped_classes['Wall']
    Zone = mapped_classes['Zone']
    TemperatureBoundaryCondition = mapped_classes['TemperatureBoundaryCondition']
    HeatFluxBoundaryCondition = mapped_classes['HeatFluxBoundaryCondition']
    Source = mapped_classes['Source']
    Assembly = mapped_classes['Assembly']
    ReferenceFace = mapped_classes['ReferenceFace']
    Volume = mapped_classes['Volume']
    Mesh = mapped_classes['Mesh']

    t_out_1 = TemperatureBoundaryCondition(name='t_out_1',
                                           temperature=273.15,
                                           data_model=data_model)

    t_out_2 = TemperatureBoundaryCondition(name='t_out_2',
                                           temperature=283.15,
                                           data_model=data_model)

    concrete = Material(name='concrete',
                        thermal_conductivity=1.5,
                        density=2000,
                        specific_heat=800,
                        data_model=data_model)

    insulation = Material(name='insulation',
                          thermal_conductivity=0.03,
                          density=150,
                          specific_heat=1500,
                          data_model=data_model)

    plaster = Material(name='plaster',
                       thermal_conductivity=0.6,
                       density=1200,
                       specific_heat=700,
                       data_model=data_model)

    construction_1 = Construction(name='construction_1',
                                  layers=[WallLayer(name='concrete_layer', thickness=0.2, material=concrete),
                                          WallLayer(name='insulation_layer', thickness=0.1, material=insulation),
                                          WallLayer(name='plaster_layer', thickness=0.01, material=plaster)],
                                  data_model=data_model)

    construction_2 = Construction(name='construction_2',
                                  layers=[WallLayer(name='concrete_layer', thickness=0.15, material=concrete),
                                          WallLayer(name='insulation_layer', thickness=0.2, material=insulation)
                                          ],
                                  data_model=data_model
                                  )

    wall1 = Wall(name='Wall1',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_1,
                 boundary_condition=t_out_1,
                 data_model=data_model)
    wall1.associate(cube.faces[0])

    wall2 = Wall(name='Wall2',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_2,
                 data_model=data_model)
    wall2.associate(cube.faces[1])

    wall3 = Wall(name='Wall3',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_1,
                 boundary_condition=t_out_1,
                 data_model=data_model)
    wall3.associate(cube.faces[2])

    wall4 = Wall(name='Wall4',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_2,
                 data_model=data_model)
    wall4.associate(cube.faces[3])

    wall5 = Wall(name='Wall5',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_1,
                 data_model=data_model)
    wall5.associate(cube.faces[4])

    wall6 = Wall(name='Wall6',
                 r_si=0.13,
                 r_se=0.04,
                 construction=construction_2,
                 boundary_condition=t_out_2,
                 data_model=data_model)
    wall6.associate(cube.faces[5])

    convective_heat_source = Source(name='convective_heat_source',
                                    heat_flux=1000,
                                    data_model=data_model)

    zone1 = Zone(name='zone_1',
                 walls=[wall1, wall2, wall3, wall4, wall5, wall6],
                 sources=[convective_heat_source],
                 data_model=data_model)
    zone1.associate(cube)

    print(cube.components)
    print(list(zone1.associated_geometry))

    steady_state_temperature = zone1.calculate_steady_state_temperature()
    print(steady_state_temperature)


def init_project(data_model: DataModel = None) -> dict[str, type]:

    classes = create_classes()
    mapped_classes = create_mapped_classes(classes)

    TypeTestClass = mapped_classes['TypeTestClass']
    Material = mapped_classes['Material']
    WallLayer = mapped_classes['Layer']
    Construction = mapped_classes['Construction']
    Wall = mapped_classes['Wall']
    Zone = mapped_classes['Zone']
    TemperatureBoundaryCondition = mapped_classes['TemperatureBoundaryCondition']
    HeatFluxBoundaryCondition = mapped_classes['HeatFluxBoundaryCondition']
    Source = mapped_classes['Source']
    Assembly = mapped_classes['Assembly']
    ReferenceFace = mapped_classes['ReferenceFace']
    Volume = mapped_classes['Volume']
    Mesh = mapped_classes['Mesh']

    type_test = TypeTestClass(name='type_test',
                              int_value=1,
                              float_value=1.0,
                              str_value='test',
                              bool_value=True,
                              list_value=[Material(name='material 1'),
                                          Material(name='material 2'),
                                          Material(name='material 3')],
                              dict_value={'a': 1, 'b': 2},
                              none_value=None,
                              obj_value=Material(name='material'),
                              numpy_value=np.array([
                                  [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                                  [[10, 11, 12], [13, 14, 15], [16, 17, 18]],
                                  [[19, 20, 21], [22, 23, 24], [25, 26, 27]],
                                  [[28, 29, 30], [31, 32, 33], [34, 35, 36]]
                              ]),
                              pandas_value=pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]}),
                              file_value=None)

    create_new_steady_state_zone(mapped_classes,
                                 data_model=data_model)

    data_model.save()
    data_model.cleanup()
    mapper.clear()

    return mapped_classes


def map_methods(mapped_classes):

    def create_asset(*args, **kwargs):

        data_model = kwargs.get('data_model')

        new_asset = FileInfo.from_string(filename='new_asset.txt',
                                         data_model=data_model,
                                         content='This is a new asset file')

        print('Asset created')

    cls = mapper.get_mapped_class('Zone')

    method_mapper.register_method(cls=cls,
                                  name='calculate_steady_state_temperature',
                                  method=cls.calculate_steady_state_temperature,
                                  args=[],
                                  kwargs={})

    method_mapper.register_method(cls=cls,
                                  name='calculate_area',
                                  method=cls.calculate_area,
                                  args=[],
                                  kwargs={})

    method_mapper.register_method(name='Create new steady state zone',
                                  method=create_new_steady_state_zone,
                                  args=[mapped_classes],
                                  add_data_model_to_kwargs=True)

    method_mapper.register_method(name='create_asset',
                                  method=create_asset,
                                  add_data_model_to_kwargs=True)

    method_mapper.register_method(name='Say hello',
                                  method=lambda: print('Hello!'),
                                  args=[],
                                  kwargs={})


mapper.clear()

mapped_classes = init_project(data_model=new_data_model)
map_methods(mapped_classes)
run_ui()

print('Test passed 48 47 2')



import numpy as np
import pandas as pd
from typing import Type
from nicegui import Client, app, ui, events
from .type_view import TypeView
from .type_view_manager import TypeViewManager

# from .detail_views import show_detail
from .. import user_manager
from ..core.edit_dialog import ContentEditDialog

from SIMULTAN.Data.MultiValues import (SimMultiValueField3D, SimMultiValueField3DParameterSource, SimMultiValueBigTable,
                                       SimMultiValueBigTableHeader, SimMultiValueBigTableParameterSource)

from PySimultan2.multi_values import simultan_multi_value_field_3d_to_numpy


class NDArrayDetailView(object):

    def __init__(self, *args, **kwargs) -> None:
        self.component: SimMultiValueField3D = kwargs.get('component')
        self.parent = kwargs.get('parent')
        self.array = None
        self.dim_slider = None
        self.table = None

    def ui_content(self, *args, **kwargs):

        from .detail_views import show_detail

        with ui.row().classes('w-full'):
            ui.input(label='Name',
                     value=self.component.Name).classes('w-full').bind_value(self.component, 'Name')
        with ui.row().classes('w-full'):
            ui.label('ID:').classes('w-full')
            with ui.row():
                with ui.row():
                    ui.label(f'{self.component.Id.GlobalId.ToString()}')
                with ui.row():
                    ui.label(f'{self.component.Id.LocalId}')

        self.array = simultan_multi_value_field_3d_to_numpy(self.component)

        with ui.row().classes('w-full'):
            with ui.column():
                ui.label('Shape:')
                ui.label(f'{self.array.shape}')

        ui.separator()

        self.table_ui_content()

        with ui.card().classes('w-full h-full'):
            ui.label('Dimension to display:')
            self.dim_slider = ui.slider(min=0, max=self.array.shape[0] - 1,
                                        step=1,
                                        value=0,
                                        on_change=self.table_ui_content.refresh)
            ui.input(value='0').bind_value(self.dim_slider,
                                           'value',
                                           forward=lambda x: int(x),
                                           backward=lambda x: str(x))

    @ui.refreshable
    def table_ui_content(self):

        if self.array.shape.__len__() > 2:
            disp_array = self.array[int(self.dim_slider.value if self.dim_slider is not None else 0), :, :]
        else:
            disp_array = self.array

        self.table = ui.table.from_pandas(pd.DataFrame(disp_array),
                                          ).classes('w-full h-full')
        with self.table.add_slot('top-left'):
            def toggle() -> None:
                self.table.toggle_fullscreen()
                button.props('icon=fullscreen_exit' if self.table.is_fullscreen else 'icon=fullscreen')

            button = ui.button('Toggle fullscreen', icon='fullscreen', on_click=toggle).props('flat')


class NDArrayView(TypeView):

    detail_view = NDArrayDetailView

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.content = kwargs.get('content', None)

    @ui.refreshable
    def ui_content(self):
        from .detail_views import show_detail
        with ui.card().classes(f"{self.colors['item']} w-full h-full") as self.card:
            self.card.on('click', lambda e: show_detail(value=self.component,
                                                        parent=self))
            with ui.row().classes('bg-stone-100 w-full') as self.row:
                self.row.on('click', lambda e: show_detail(value=self.component,
                                                           parent=self))
                self.checkbox = ui.checkbox(on_change=self.select)
                ui.label(f'{self.component.Id}')
                ui.label(f'{self.component.Name}')


class NDArrayManager(TypeViewManager):

    cls: np.ndarray = np.ndarray
    item_view_cls: Type[TypeView] = NDArrayView
    item_view_name = 'ND Arrays'

    def update_items(self) -> list[SimMultiValueField3D]:
        if self.data_model is None:
            return []
        return [x for x in self.data_model.value_fields if type(x) == SimMultiValueField3D]

    def button_create_ui_content(self):
        ui.button('Create new ND-Array', on_click=self.create_new_item, icon='add')

    @ui.refreshable
    def add_item_to_view(self,
                         item: any,
                         raw_val=None):

        if isinstance(item, SimMultiValueField3D):
            val_source = item
        elif isinstance(item, np.ndarray):
            val_source: SimMultiValueField3D = raw_val.ValueSource.ValueField

        if self.items_ui_element is None:
            return

        if val_source not in self.items:
            self.items.append(val_source)
        item_view = self.item_views.get(str(val_source.Id), None)

        if item_view is None:
            item_view = self.item_view_cls(component=val_source,
                                           parent=self)
            self.item_views[str(val_source.Id)] = item_view
            with self.items_ui_element:
                item_view.ui_content()
        else:
            if item_view.card.parent_slot.parent.parent_slot.parent is not self.items_ui_element:
                with self.items_ui_element:
                    item_view.ui_content()
        return item_view



import os
import asyncio
from copy import deepcopy
from ..core import mapper
from logging import getLogger
from typing import Optional, Type
from datetime import datetime
from nicegui import app, ui, events
from PySimultan2.data_model import DataModel
from PySimultan2.object_mapper import PythonMapper

from .. import user_manager
from ..core.user import UserManager, User
from ..views.view_manager import ViewManager
from ..views.asset_view import AssetManager
from ..views.geometry_view import GeometryManager
from ..core.navigation import Navigation
# from ..views.geometry_view.geometry_file_manager import GeometryFileManager
# from ..views.asset_view.asset_manager import AssetManager

from ..app.project import content as project_content

import shutil


logger = getLogger('py_simultan_ui')


project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)

app.add_static_files('/project', project_dir)


class NewProjectDialog(object):

    def __init__(self, *args, **kwargs):
        self.dialog = None
        self.parent = kwargs.get('parent', None)

    def validate_project_name(self, project_name):
        if not self.dialog.project_name_input.value.endswith('.simultan'):
            return "Project name must end with '.simultan'!"

    def ui_content(self):
        with ui.dialog() as dialog, ui.card():
            self.dialog = dialog
            with ui.row():
                project_name_input = ui.input('Project name',
                                              #validation=self.validate_project_name
                                              )
                ui.label('.simultan')

            user_name_input = ui.input('User name',
                                       value='admin',
                                       # validation=self.validate_project_name
                                       )
            password_input = ui.input('Password',
                                      value='admin',
                                      # validation=self.validate_project_name
                                      )

            dialog.project_name_input = project_name_input
            dialog.user_name_input = user_name_input
            dialog.password_input = password_input

            with ui.row():
                create_btn = ui.button('Create', on_click=self.new_project)
                ui.button('Cancel', on_click=dialog.close)

    def new_project(self, e: events.ClickEventArguments):
        self.dialog.close()

        project_path = f'{project_dir}/{self.dialog.project_name_input.value}.simultan'
        data_model = DataModel.create_new_project(project_path=project_path,
                                                  user_name=self.dialog.user_name_input.value,
                                                  password=self.dialog.password_input.value)
        data_model.save()
        data_model.cleanup()
        if isinstance(data_model, DataModel):
            ui.notify(f'Project {self.dialog.project_name_input.value} created!')
            self.parent.project_list.ui_content.refresh()
        else:
            ui.notify(f'Error creating project {self.dialog.project_name_input.value}!')


class ProjectView(object):

    def __init__(self, *args, **kwargs):

        self._selected: bool = False

        self.parent: Optional[ProjectList] = kwargs.get('parent', None)
        self.project: Optional[str] = kwargs.get('project', None)
        self.project_manager: Optional[ProjectManager] = kwargs.get('project_manager', None)

        self.checkbox: Optional[ui.checkbox] = None
        self.card: Optional[ui.item] = None

        self.selected = kwargs.get('selected', False)

    @property
    def selected(self) -> bool:
        return self._selected

    @selected.setter
    def selected(self, value: bool):
        self._selected = value
        self.ui_content.refresh()

    @property
    def size(self):
        return os.path.getsize(f'{project_dir}/{self.project}')

    @property
    def last_modified(self):
        return datetime.fromtimestamp(os.path.getmtime(f'{project_dir}/{self.project}'))

    @property
    def path(self):
        return f'{project_dir}/{self.project}'

    @property
    def project_dict(self):
        return {'size': self.size,
                'last_modified': self.last_modified,
                'path': self.path,
                'project': self.project,
                }

    @ui.refreshable
    def ui_content(self):
        with ui.item() as self.card:
            with ui.item_section():
                ui.label(self.project).classes('text-xl')
            with ui.item_section():
                if self.size < 1024:
                    ui.label(f'Size: {self.size:.2f} B')
                elif self.size < 1024 ** 2:
                    ui.label(f'Size: {self.size / 1024:.2f} kB')
                else:
                    ui.label(f'Size: {self.size / 1024 ** 2:.2f} MB')
            with ui.item_section():
                ui.label(f'Last modified: {self.last_modified}')

            if self.project.endswith('.simultan') and not self.project.startswith('~'):
                if self.selected:
                    with ui.item_section():
                        self.card.classes('bg-blue-5')
                        sel_button = ui.button('Close',
                                               icon='close',
                                               on_click=self.parent.close_project).classes('q-ml-auto')
                        sel_button.project = self
                        sel_button.project_item = self.card
                    with ui.item_section():
                        save_button = ui.button('Save',
                                                icon='save',
                                                on_click=self.parent.save_project).classes('q-ml-auto')
                        save_button.project = self

                else:
                    with ui.item_section():
                        sel_button = ui.button('Open',
                                               icon='file_open',
                                               on_click=self.parent.select_project).classes('q-ml-auto')
                        sel_button.project = self
                        sel_button.project_item = self.card
                    with ui.item_section():
                        dl_button = ui.button(icon='download',
                                              on_click=self.parent.download_project).classes('q-ml-auto')
                        dl_button.project = self

                    with ui.item_section():
                        del_button = ui.button(icon='delete',
                                               on_click=self.parent.delete_project).classes('q-ml-auto')
                        del_button.project = self
            else:
                with ui.item_section():
                    pass
                with ui.item_section():
                    dl_button = ui.button(icon='download',
                                          on_click=self.download_project).classes('q-ml-auto')
                    dl_button.project = self.project_dict
                with ui.item_section():
                    del_button = ui.button(icon='delete',
                                           on_click=self.parent.delete_project).classes('q-ml-auto')
                    del_button.project = self

    def download_project(self, e: events.ClickEventArguments):
        ui.download(f'project/{self.project}')


class ProjectList(object):

    def __init__(self, *args, **kwargs):
        self._projects = None
        self.selected_project = kwargs.get('selected_project', None)
        self.project_manager = kwargs.get('project_manager', None)
        self.project_views = kwargs.get('project_views', {})

        self.file_list = None

    @property
    def projects(self):
        if self._projects is None:
            self.projects = self.refresh_projects()
        return self._projects

    @projects.setter
    def projects(self, value):

        old_set = set(self._projects) if self._projects is not None else set()
        new_set = set(value) if value is not None else set()
        self._projects = value

        common_projects = old_set.intersection(new_set)
        new_projects = new_set.difference(common_projects)
        removed_projects = old_set.difference(common_projects)

        for project in new_projects:
            self.add_project_to_view(project)

        for project in removed_projects:
            if project in self.project_views:
                self.project_views[project].card.parent_slot.parent.remove(self.project_views[project].card)

    def refresh_projects(self):
        return os.listdir(project_dir)

    @property
    def selected_project(self):
        return app.storage.user['selected_project']

    @selected_project.setter
    def selected_project(self, value):
        app.storage.user['selected_project'] = value

    @ui.refreshable
    def ui_content(self):
        self.project_views = {}
        ui.label('Files:').classes('text-2xl')
        with ui.list().classes('w-full h-full') as self.file_list:
            for project in os.listdir(project_dir):
                project_view = ProjectView(project=project,
                                           selected=project == self.selected_project,
                                           project_manager=self.project_manager,
                                           parent=self)
                self.project_views[project] = project_view
                project_view.ui_content()

    def add_project_to_view(self, project: str):
        if project not in self.project_views:
            project_view = ProjectView(project=project,
                                       selected=project == self.selected_project,
                                       project_manager=self.project_manager,
                                       parent=self)
            self.project_views[project] = project_view
            with self.file_list:
                project_view.ui_content()

    async def select_project(self, e: events.ClickEventArguments):
        self.selected_project = e.sender.project.project
        if self.project_manager is not None:
            await self.project_manager.open_project(e)
        self.projects = self.refresh_projects()

    def close_project(self, e: events.ClickEventArguments):
        self.selected_project = None
        self.project_manager.close_project(e)
        self.projects = self.refresh_projects()

    def delete_project(self, e: events.ClickEventArguments):
        if os.path.isfile(e.sender.project.path):
            file = e.sender.project.path
            os.remove(file)
            ui.notify(f"Project {e.sender.project.project} deleted!")
            self.ui_content.refresh()
        elif os.path.isdir(e.sender.project.path):
            shutil.rmtree(e.sender.project.path)
            ui.notify(f"Project {e.sender.project.project} deleted!")
            self.ui_content.refresh()

    def save_project(self, e: events.ClickEventArguments):
        if self.project_manager.data_model is not None:
            self.project_manager.data_model.save()
            ui.notify(f"Project {e.sender.project.project} saved!")
        else:
            ui.notify(f"Project {e.sender.project.project} not saved! No project loaded!")

    def download_project(self, e: events.ClickEventArguments):
        ui.download(f'project/{e.sender.project.project}')


class ProjectManager(object):
    def __init__(self, *args, **kwargs):

        self.user_manager: UserManager = kwargs.get('user_manager', user_manager)
        self._projects = None
        self.project_list = None
        self.mapped_data = None

        if kwargs.get('data_model', None) is not None:
            self.data_model = kwargs.get('data_model', None)

    @property
    def mapper(self) -> PythonMapper:
        return self.user.mapper

    @mapper.setter
    def mapper(self, value: PythonMapper):
        self.user.mapper = value

    @property
    def grid_view(self):
        return self.user.grid_view

    @property
    def view_manager(self) -> ViewManager:
        return self.user.view_manager

    @property
    def asset_manager(self) -> AssetManager:
        return self.user.asset_manager

    @property
    def geometry_manager(self) -> GeometryManager:
        return self.user.geometry_manager

    @property
    def array_manager(self):
        return self.user.array_manager

    @property
    def navigation(self) -> Navigation:
        return self.user.navigation

    @property
    def detail_view(self):
        return self.user.detail_view

    @property
    def home_tab(self):
        return self.user.home_tab

    @property
    def project_tab(self):
        return self.user.project_tab

    @property
    def user(self) -> User:
        return self.user_manager.users[app.storage.user['username']]

    @property
    def logger(self):
        return self.user.logger

    @property
    def data_model(self) -> Optional[DataModel]:
        return self.user.data_model

    @data_model.setter
    def data_model(self, value: Optional[DataModel]):
        logger.debug(f'Setting data model to {value}')
        self.user.data_model = value
        if self.user.data_model is not None:
            app.add_static_files('/assets', self.user.data_model.project.ProjectUnpackFolder.FullPath)

        for item in [self.view_manager, self.asset_manager, self.geometry_manager, self.grid_view, self.array_manager]:
            if item is not None:
                item.data_model = self.user.data_model

    @property
    def projects(self):
        # get list of files in project_dir
        self._projects = os.listdir(project_dir)
        return self._projects

    @property
    def selected_project(self):
        return self.project_list.selected_project

    def upload_project(self,
                       e: events.UploadEventArguments,
                       *args,
                       **kwargs):

        shutil.copyfileobj(e.content, open(f'{project_dir}/{e.name}', 'wb'))
        ui.notify(f'Project {e.name} uploaded!')
        self.project_list.ui_content.refresh()

    def refresh_all_items(self, *args, **kwargs):
        # self.view_manager.refresh_all_items()
        self.grid_view.ui_content.refresh()
        self.asset_manager.update_items_views()
        self.geometry_manager.update_items_views()
        # self.navigation.ui_content.refresh()

    def ui_content(self):

        self.user.add_ui_log()

        with self.project_tab:
            with ui.tabs() as tabs:
                ui.tab('Components', icon='category', label='')
                ui.tab('Assets', icon='description', label='')
                ui.tab('Geometry', icon='shape_line', label='')
                ui.tab('Arrays', icon='data_array', label='')

            with ui.tab_panels(tabs, value='Components').classes('w-full h-full').props('vertical'):
                with ui.tab_panel('Components').classes('w-full h-full'):
                    self.grid_view.ui_content()
                # self.view_manager.ui_content()
                with ui.tab_panel('Assets').classes('w-full h-full'):
                    self.asset_manager.ui_content()
                with ui.tab_panel('Geometry').classes('w-full h-full'):
                    self.geometry_manager.ui_content()

                with ui.tab_panel('Arrays').classes('w-full h-full'):
                    self.array_manager.ui_content()

            # self.navigation.ui_content()

        if self.data_model is None:
            selected = None
        else:
            selected = os.path.basename(self.data_model.project.ProjectFile.FullPath)

        self.project_list = ProjectList(project_manager=self,
                                        selected_project=selected
                                        )
        self.project_list.ui_content()

        new_project_dialog = NewProjectDialog(parent=self)
        new_project_dialog.ui_content()

        ui.button('New project', on_click=new_project_dialog.dialog.open).classes('max-w-full')

        ui.upload(label='Upload simultan project',
                  on_upload=self.upload_project).on(
            'finish', lambda: ui.notify('Finish!')
        ).classes('max-w-full')

    async def open_project(self, e: events.ClickEventArguments = None, *args, **kwargs):

        if e is not None:
            project = e.sender.project
        else:
            project = kwargs.get('project', None)

        # dialog to ask username and password and if to load all data
        with ui.dialog() as dialog, ui.card():

            ui.label(f'Open project: {project.project}')

            with ui.row():
                user_input = ui.input('User name', value='admin')
                pw_input = ui.input('Password', value='admin', password=True)

            open_all = ui.checkbox('Load all data', value=True)

            with ui.row():
                ui.button('OK', on_click=lambda: dialog.submit({'ok': True,
                                                                 'user': user_input.value,
                                                                 'pw': pw_input.value,
                                                                 'open_all': open_all.value}
                                                                )
                          )
                ui.button('Cancel', on_click=lambda: dialog.submit({'ok': False}))

        result = await dialog

        self.user.logger.info(f'Opening project {project.project}')

        if result is None or not result.get('ok', False):
            return

        self.mapper = self.user.mapper
        self.mapper.clear()

        logger.info(f'Opening project {project.project}')

        n = ui.notification(timeout=None)
        n.spinner = True
        n.message = f'Opening project {project.project}'

        project_dict = project.project_dict
        new_data_model = DataModel(project_path=project_dict['path'],
                                   user_name=result['user'],
                                   password=result['pw'])
        self.mapped_data = new_data_model.get_typed_data(self.mapper, create_all=True if result['open_all'] else False)
        self.data_model = new_data_model
        project.selected = True
        self.project_list.ui_content.refresh()
        n.message = f'Project {project.project} opened!'
        n.spinner = False
        n.dismiss()

    def open_data_model(self,
                        data_model: DataModel = None,
                        create_all: bool = True):
        self.mapper.clear()
        if data_model is None:
            data_model = self.data_model

        if data_model is not None:
            self.mapped_data = data_model.get_typed_data(self.mapper,
                                                         create_all=create_all)
        self.data_model = data_model

        ui.notify('Data model loaded!')

    def close_project(self, e):

        from ..views.detail_views import show_detail

        self.logger.info(f'Closing project {os.path.basename(self.data_model.project.ProjectFile.FullPath)}')

        if self.data_model is not None:
            self.data_model.cleanup()
        self.mapper.clear()
        self.mapped_data = []
        self.data_model = None
        ui.notify('Project closed!')
        self.project_list.ui_content.refresh()
        e.sender.project.selected = False
        self.project_list.ui_content.refresh()
        show_detail(None)

        # self.geometry_file_manager.geometry_models = []
        # project_content.refresh()



import logging
from nicegui import ui, app
from .. import user_manager


class LogElementHandler(logging.Handler):
    """A logging handler that emits messages to a log element."""

    def __init__(self, element: ui.log, level: int = logging.NOTSET) -> None:
        self.element = element
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            self.element.push(msg)
        except Exception:
            self.handleError(record)


class UILog:
    def __init__(self):
        self.log = None
        self.handler = None
        self.logger = logging.getLogger(self.user.name)

    @property
    def user(self):
        return user_manager.users[app.storage.user['username']]

    def ui_content(self):
        self.log = ui.log(max_lines=100).classes('w-full h-full')
        self.log.style = "font-family: monospace; white-space: pre-wrap;"
        self.handler = LogElementHandler(self.log)

        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

        for logger in loggers:
            logger.addHandler(self.handler)

        self.logger = logging.getLogger(self.user.name)
        self.logger.addHandler(self.handler)
        self.logger.setLevel(logging.DEBUG)

    def __del__(self):
        self.logger.removeHandler(self.handler)
        for logger in logging.root.manager.loggerDict.values():
            try:
                logger.removeHandler(self.handler)
            except:
                pass

    def clear(self):
        self.log.clear()



class Class2:
    def __init__(self, *args, **kwargs):
        self.param_a = kwargs.get('param_a', None)
        self.param_b = kwargs.get('param_b', None)
        self.param_c = kwargs.get('param_c', None)
        self.param_d = kwargs.get('param_d', None)

    def multiply(self):
        self.param_d = self.param_a * self.param_b


class Class3:
    def __init__(self, *args, **kwargs):
        self.thermal_conductivity = kwargs.get('thermal_conductivity', 15)
        self.thermal_resistance = kwargs.get('thermal_resistance', 30)

    def test_method(self):
        print(f'This is a test method {self}')
